# External
from os   import environ
# Internal
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.rule import Rule
from dbgen.support.datatypes.temp      import Temp,TempFunc
from dbgen.support.datatypes.misc      import Arg
from dbgen.support.utils               import mkInsCmd,mkUpdateCmd
from dbgen.support.datatypes.metatemp  import MetaTemp,Block,SimpleUpdate,PureSQL
from dbgen.support.datatypes.sqltypes  import Varchar
from dbgen.inputs.extra                import surface

from dbgen.scripts.Pure.Atoms.get_bulk           import get_bulk
from dbgen.scripts.Pure.Atoms.get_miller         import get_miller
from dbgen.scripts.Pure.Atoms.surface_bulkstruct import surface_bulkstruct
from dbgen.scripts.Pure.Atoms.get_pure_struct    import get_pure_struct
from dbgen.scripts.Pure.Atoms.json_to_traj       import json_to_traj
from dbgen.scripts.Pure.Atoms.get_miller         import get_miller
from dbgen.scripts.Pure.Misc.pmg_to_ase          import pmg_to_ase
################################################################
################################################################
pypth = environ['DBGEN_ROOT']+'/scripts/'

functions = {
     'get_bulk'             : pypth + "Pure/Atoms/get_bulk.py"
    ,'get_miller'           : pypth + "Pure/Atoms/get_miller.py"
    ,'surface_bulkstruct'   : pypth + "Pure/Atoms/surface_bulkstruct.py"
    ,'get_pure_struct'      : pypth + "Pure/Atoms/get_pure_struct.py"
    ,'json_to_traj'         : pypth + "Pure/Atoms/json_to_traj.py"
    ,'pmg_to_ase'           : pypth + "Pure/Misc/pmg_to_ase.py"}
################################################################
################################################################
################################################################

species = Table('species'
    ,desc="""A combination of composition and structural information (no floats)"""
    ,cols=[Col('species_id',               pk   = True, auto = True)
          ,Col('name',          Varchar(), uniq = True)
          ,Col('nickname',      Varchar())
          ,Col('composition',   Varchar())])

bulkspecies = Table('bulkspecies'
    ,desc = 'Properties of species that are bulks'
    ,cols = [Col('species_id',    pk=True)
            ,Col('pure_struct_id',nn=True)]
    ,fks  = [FK('species_id',    'species')
            ,FK('pure_struct_id','pure_struct')])

molspecies = Table('molspecies'
    ,desc = 'Properties of species that are molecules. IN THE FUTURE, the SMILES string should be here because pointgroup + composition is underspecified'
    ,cols = [Col('species_id',pk = True)
            ,Col('pointgroup',nn = True)]
    ,fks  = [FK('species_id','species')
             ,FK('pure_struct_id','pure_struct')])

surfspecies = Table('surfspecies'
    ,desc = 'How should this be specified ??? And 2D materials?'
    ,cols = [Col('species_id',                  pk = True)
            ,Col('pure_struct_id',              nn = True)
            ,Col('facet',           Varchar(),  nn = True)]
    ,fks  = [FK('species_id','species')
            ,FK('pure_struct_id','pure_struct')])

species_composition = Table('species_composition'
    ,desc='Mapping table between species and element'
    ,cols = [Col('species_id',pk = True)
            ,Col('element_id',pk = True)
            ,Col('count',     nn = True)]
    ,fks  = [FK('species_id','species')
            ,FK('element_id','element')])

########################################################

pure_struct = Table('pure_struct'
    ,desc = 'Structure abstraction developed by Ankit Jain'
    ,cols = [Col('pure_struct_id',              pk = True,auto = True)
            ,Col('name',            Varchar(),  nn = True,uniq = True)
            ,Col('spacegroup')
            ,Col('free')
            ,Col('nickname',        Varchar())
            ])

tables = [species
         ,bulkspecies
         ,molspecies
         ,surfspecies
         ,species_composition
         ,pure_struct]

################################################################
################################################################
################################################################
nick_dict = {'AB_1_a_b_225'     : 'rocksalt'
            ,'AB_1_a_c_216'     : 'zincblende'
            ,'AB_1_a_b_221'     : 'cesium chloride'
            ,'A_1_a_225'        : 'fcc'
            ,'A_1_a_229'        : 'bcc'
            ,'A_2_c_194'        : 'hcp'
            ,'A_2_a_227'        : 'diamond'
            ,'AB3_1_a_d_221'    : 'anti-ReO3'
            ,'A2B3_8_ad_e_206'  : 'antibixbyite'
            # add rutile/perovskite
            }

pop_nickname = Rule('pop_nickname'
    ,metatemp=PureSQL(["""UPDATE pure_struct
                        SET   nickname = '{0}'
                        WHERE name     = '{1}'
                        AND pure_struct_id > 0
                        """.format(v,k) for k,v in nick_dict.items()]))

pop_compstr = Rule('species_compstr'
    ,metatemp=PureSQL([""" UPDATE species AS S SET S.composition=
                  (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                        FROM species_composition as C
                        WHERE C.species_id = S.species_id)"""]))

update_species_fk = Rule('species_fk'
    ,metatemp= PureSQL(["""UPDATE struct AS S SET S.species_id = (
                SELECT  P.species_id FROM species AS P
                LEFT JOIN bulkspecies B USING (species_id)
                LEFT JOIN molspecies M USING (species_id)
                WHERE S.composition_norm = P.composition
                 AND    (S.system_type = 'bulk'
                        AND B.pure_struct_id = S.pure_struct_id)
                        OR (S.system_type = 'molecule'
                        AND M.pointgroup = S.symmetry))
                WHERE S.struct_id > 0"""]))

pop_pure = Rule('pop_pure_struct'
    ,query    = 'SELECT S.struct_id,S.raw FROM struct AS S WHERE pure_struct_id IS NULL'
    ,template = Temp(tempfuncs=[TempFunc(get_bulk
                                        ,args=[Arg('json_to_traj')])
                                ,TempFunc(json_to_traj
                                        ,args=[Arg('raw')])
                                ,TempFunc(get_pure_struct
                                        ,args=[Arg('get_bulk')])])
    ,metatemp=MetaTemp([Block(name = 'insert'
                             ,text = mkInsCmd('pure_struct',['name','spacegroup','free'],sqlite=False)
                             ,args = [Arg('get_pure_struct',i) for i in range(3)])
                        ,Block(name='query'
                            ,text="""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                            ,args=[Arg('get_pure_struct',0)])
                        ,Block(name='insert_fk'
                                ,text=mkUpdateCmd('struct',['pure_struct_id'],['struct_id'])
                                ,args=[Arg('query',0),Arg('struct_id')])]))

underlying = Rule('underlying'
    ,query    = """SELECT U.struct_id,S.raw
                    FROM struct AS S
                    JOIN surface AS U USING (struct_id)
                    WHERE U.facet_h IS NULL"""
    ,template = Temp(tempfuncs = [TempFunc(json_to_traj
                                          ,args = [Arg('raw')])
                                 ,TempFunc(surface_bulkstruct
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(pmg_to_ase
                                         ,args=[Arg('surface_bulkstruct')])
                                 ,TempFunc(get_bulk
                                          ,args=[Arg('pmg_to_ase')])
                                 ,TempFunc(get_miller
                                          ,args=[Arg('surface_bulkstruct')])
                                 ,TempFunc(get_pure_struct
                                          ,args=[Arg('get_bulk')])])
    ,metatemp = MetaTemp([Block(name = 'insert'
                             ,text = mkInsCmd('pure_struct',['name','spacegroup','free'],sqlite=False)
                             ,args = [Arg('get_pure_struct',i) for i in range(3)])
                        ,Block(name='query'
                            ,text="""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                            ,args=[Arg('get_pure_struct',0)])
                        ,Block(name='insert_fk'
                                ,text=mkUpdateCmd('surface',['pure_struct_id'],['struct_id'])
                                ,args=[Arg('query',0),Arg('struct_id')])
                        ,Block(name='update_miller'
                              ,text = mkUpdateCmd('surface',['facet_h','facet_k','facet_l'],['struct_id'])
                              ,args = [Arg('get_miller',i) for i in range(3)]
                                     +[Arg('struct_id')])]))

rules = [pop_pure
             ,underlying
             ,pop_nickname
             ,pop_compstr
             ,update_species_fk]



# External Modules
from typing import  Iterator,Tuple,Set,List,Any,Dict
from sqlparse        import parse             # type: ignore
from sqlparse.sql    import Token             # type: ignore
from sqlparse.tokens import Name,Whitespace   # type: ignore
from abc             import abstractmethod
"""
Functions to assist with the extraction of dependencies from a raw SQL query
"""

class Stmt(object):
    pass

    @abstractmethod
    def parents(self)->Tuple[Set[str],Set[str]]:
        """
        What tables / columns does this statement depend on existing
        """
        raise NotImplementedError

    @abstractmethod
    def children(self)->Tuple[Set[str],Set[str]]:
        """
        What tables (i.e. insertions) / columns does this statement modify?
        """
        raise NotImplementedError

class SelectStmt(Stmt):
    """
    Collection of tokens associated with any SELECT statement
    """
    def __init__(self,tokens:list,alias:dict={})->None:
        eof = [Token(None,'<EOF>')]
        self.tokens = [t for t in tokens if t.ttype != Whitespace] + eof
        self.alias  = alias # type: Dict[str,str]

    def parents(self)->Tuple[Set[str],Set[str]]:
        tabdata,coldata    = set(),set() # type: Tuple[Set[str],Set[Tuple[str,str]]]
        last_tok,penult = '','' # initialize placeholders
        for i,tok in enumerate(self.tokens):
            t = tok.normalized
            if t == '<EOF>': break
            if tok.ttype == Name:
                if self.tokens[i+1].normalized=='.':
                    coldata.add((t,self.tokens[i+2].normalized))
                elif self.tokens[i+1].normalized=='AS':
                    self.alias[self.tokens[i+2].normalized] = t
                    tabdata.add(t)
        cols = set(['%s.%s'%(self.alias.get(t,t),c) for t,c in coldata])
        tabs = tabdata
        return tabs,cols

    def children(self)->Tuple[Set[str],Set[str]]:
        return set(),set()

class InsertStmt(Stmt):
    """
    Any INSERT statment of the form "INSERT INTO <tab> ([cols]) VALUES (...)"
    """
    def __init__(self,tabname:str,inscols:list,insvals:list)->None:
        self.tabname = tabname
        self.inscols = inscols
        self.insvals = insvals
    def parents(self)->Tuple[Set[str],Set[str]]:
        return SelectStmt(self.insvals).parents()
    def children(self)->Tuple[Set[str],Set[str]]:
        cols = ['%s.%s'%(self.tabname,c) for c in self.inscols]
        return set([self.tabname]),set(cols)

class UpdateStmt(Stmt):
    """
    Any UPDATE statment of the form "UPDATE <tab> [AS alias] SET [colval_pairs]"
    """
    def __init__(self,tabname:str,colvals:Iterator,alias:dict)->None:
        self.tabname = tabname
        self.colvals = colvals
        self.alias   = alias
        #print('len colvals  = %d'%len(colvals))
    def parents(self)->Tuple[Set[str],Set[str]]:
        t,c = set(),set()
        for cv in self.colvals:
            i = 4 if self.alias else 2
            pseudoSelect = SelectStmt(cv[i:],self.alias)
            newT,newC = pseudoSelect.parents()
            t.update(newT);c.update(newC)
        return t,c
    def children(self)->Tuple[Set[str],Set[str]]:
        output = set()
        for cv in self.colvals:
            i = 2 if self.alias else 0
            output.add(self.tabname+'.'+cv[i].value)
        #if any(['timestamp' in [x.value for x in cv] for cv in self.colvals]):
        #    import pdb;pdb.set_trace()
        return set(),output

def mkStmt(s:str)->Stmt:
    """
    Convert a Raw SQL string into a Stmt
    """
    #print('PARSING '+s)
    debug = 'UPDATE adsorbate' in s
    #if debug: print('PARSING '+s)
    try:
        toks = parse(s.strip())[0].flatten()
    except IndexError:
        raise ValueError('Could not parse query:\n',s)

    fst = next(toks).normalized
    next(toks)                      # whitespace
    if fst == 'SELECT':
        return SelectStmt(list(toks))

    elif fst == 'INSERT':
        assert next(toks).normalized == 'INTO';next(toks)
        tab = next(toks);next(toks)
        assert next(toks).normalized == '('
        inscols = []
        for t in toks:
            if t.normalized==')': break
            if t.normalized!=',':
                inscols.append(t.value)
        return InsertStmt(tab.value,inscols,list(toks))

    elif fst == 'UPDATE':
        tab = next(toks);next(toks)
        if next(toks).normalized=='AS':
            next(toks)
            alias = {next(toks).normalized:tab}
            next(toks);next(toks)
        else:
            alias = {};next(toks)

        colvals = [[]] # type: list
        l_toks = [t for t in toks if t.ttype !=Whitespace]
        for i,t in enumerate(l_toks):
            if t.normalized == ',' and l_toks[i+1].ttype==Name and l_toks[i+2].value=='=':
                colvals.append([])
            else : colvals[-1].append(t)
        return UpdateStmt(tab.value,list(filter(None,colvals)),alias)
    else:
        raise NotImplementedError

def map_parser(xs : List[str]) -> Tuple[Set[str],Set[str],Set[str],Set[str]]:
    """
    For many SQL statements, create Stmt objects and aggregate dependency info
    """
    pt,pc,ct,cc = set(),set(),set(),set()
    for stmt in map(mkStmt,xs):
        npt,npc = stmt.parents()
        nct,ncc = stmt.children()
        pt.update(npt);pc.update(npc)
        ct.update(nct);cc.update(ncc)
    return pt,pc,ct,cc

################################################################################
# Exported Functions
#-------------------
def parse_dependencies(xs:List[str])->Tuple[Set[str],Set[str],Set[str],Set[str]]:
    """
    List of tables,columns that are queried by the SQL statements
    """
    alltabs,allcols,deptabs,depcols =  map_parser(xs)
    return (alltabs - deptabs,allcols - depcols,deptabs,depcols)



# External Modules
from os import environ
# Internal Modules
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.temp      import Temp,TempFunc,Const,SimpleTemp
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.metatemp  import PureSQL,SimpleUpdate,MetaTemp,Block,SimpleInsert
from dbgen.support.datatypes.rule      import Rule
from dbgen.inputs.basics               import struct,traj,atom
from dbgen.core.lists                  import flatten,merge_dicts
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal

from dbgen.scripts.Pure.Load.get_dipol_gpaw         import get_dipol_gpaw
from dbgen.scripts.Pure.Load.get_dipol_qe           import get_dipol_qe
from dbgen.scripts.Pure.Load.get_dipol_vasp         import get_dipol_vasp
from dbgen.scripts.Pure.Load.get_gamma_gpaw         import get_gamma_gpaw
from dbgen.scripts.Pure.Load.get_gamma_qe           import get_gamma_qe
from dbgen.scripts.Pure.Load.get_gamma_vasp         import get_gamma_vasp
from dbgen.scripts.Pure.Load.get_econv_gpaw         import get_econv_gpaw
from dbgen.scripts.Pure.Load.get_econv_qe           import get_econv_qe
from dbgen.scripts.Pure.Load.get_econv_vasp         import get_econv_vasp
from dbgen.scripts.Pure.Load.get_sigma_gpaw         import get_sigma_gpaw
from dbgen.scripts.Pure.Load.get_sigma_qe           import get_sigma_qe
from dbgen.scripts.Pure.Load.get_sigma_vasp         import get_sigma_vasp
from dbgen.scripts.Pure.Load.get_mixing_gpaw        import get_mixing_gpaw
from dbgen.scripts.Pure.Load.get_mixing_qe          import get_mixing_qe
from dbgen.scripts.Pure.Load.get_mixing_vasp        import get_mixing_vasp
from dbgen.scripts.Pure.Load.get_nbands_gpaw        import get_nbands_gpaw
from dbgen.scripts.Pure.Load.get_nbands_qe          import get_nbands_qe
from dbgen.scripts.Pure.Load.get_nbands_vasp        import get_nbands_vasp
from dbgen.scripts.Pure.Load.get_nmix_gpaw          import get_nmix_gpaw
from dbgen.scripts.Pure.Load.get_nmix_qe            import get_nmix_qe
from dbgen.scripts.Pure.Load.get_nmix_vasp          import get_nmix_vasp
from dbgen.scripts.Pure.Load.get_diag_gpaw          import get_diag_gpaw
from dbgen.scripts.Pure.Load.get_diag_qe            import get_diag_qe
from dbgen.scripts.Pure.Load.get_diag_vasp          import get_diag_vasp
from dbgen.scripts.Pure.Load.get_spinpol_gpaw       import get_spinpol_gpaw
from dbgen.scripts.Pure.Load.get_spinpol_qe         import get_spinpol_qe
from dbgen.scripts.Pure.Load.get_spinpol_vasp       import get_spinpol_vasp
from dbgen.scripts.Pure.Load.get_dipole_gpaw        import get_dipole_gpaw
from dbgen.scripts.Pure.Load.get_dipole_qe          import get_dipole_qe
from dbgen.scripts.Pure.Load.get_dipole_vasp        import get_dipole_vasp
from dbgen.scripts.Pure.Load.get_maxstep_gpaw       import get_maxstep_gpaw
from dbgen.scripts.Pure.Load.get_maxstep_qe         import get_maxstep_qe
from dbgen.scripts.Pure.Load.get_maxstep_vasp       import get_maxstep_vasp

from dbgen.scripts.Pure.Load.get_kpts_gpaw          import get_kpts_gpaw
from dbgen.scripts.Pure.Load.get_kpts_qe            import get_kpts_qe
from dbgen.scripts.Pure.Load.get_kpts_vasp          import get_kpts_vasp
from dbgen.scripts.Pure.Misc.dftcode_dependent      import dftcode_dependent
from dbgen.scripts.Pure.Misc.identity               import identity
from dbgen.scripts.Pure.Misc.get                    import get
from dbgen.scripts.Pure.Misc.flatten                import flatten
from dbgen.scripts.Pure.Misc.map_                   import map_
from dbgen.scripts.Pure.Misc.pair                   import pair
from dbgen.scripts.Pure.Misc.triple                 import triple
from dbgen.scripts.Pure.Atoms.get_voronoi           import get_voronoi
from dbgen.scripts.Pure.Atoms.get_pointgroup        import get_pointgroup
from dbgen.scripts.Pure.Atoms.get_spacegroup        import get_spacegroup
from dbgen.scripts.Pure.Atoms.json_to_traj          import json_to_traj
from dbgen.scripts.Pure.Atoms.get_vacuum            import get_vacuum
from dbgen.scripts.Pure.Load.load_comp              import load_comp
from dbgen.scripts.Pure.Load.get_vasp_key           import get_vasp_key
from dbgen.scripts.Pure.Load.get_dw                 import get_dw
from dbgen.scripts.Pure.Load.get_cell_dofree        import get_cell_dofree
from dbgen.scripts.Pure.Load.get_cell_factor        import get_cell_factor
from dbgen.scripts.Pure.Load.get_mixtype            import get_mixtype
from dbgen.scripts.Pure.Load.get_kptden             import get_kptden
from dbgen.scripts.Pure.Graph.mk_graph              import mk_graph
from dbgen.scripts.Pure.Graph.layers                import layers
from dbgen.scripts.Pure.Graph.geometric             import geometric
from dbgen.scripts.Pure.Graph.graph_to_json         import graph_to_json
from dbgen.scripts.Pure.Graph.json_to_graph         import json_to_graph

################################################################
################################################################
################################################################
# CONSTANTS
###########
pypth = environ['DBGEN_ROOT']+'/scripts/'

dfts         = ['gpaw','qe','vasp']
other_cols = ['dipol','gamma','econv','sigma','mixing','nbands'
             ,'nmix','diag','spinpol','dipole','maxstep','kpts']
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']



get_fns = [get_dipol_gpaw,get_dipol_qe,get_dipol_vasp,get_gamma_gpaw
          ,get_gamma_qe,get_gamma_vasp,get_econv_gpaw,get_econv_qe,get_econv_vasp
          ,get_sigma_gpaw,get_sigma_qe,get_sigma_vasp,get_mixing_gpaw
          ,get_mixing_qe,get_mixing_vasp,get_nbands_gpaw,get_nbands_qe,get_nbands_vasp
          ,get_nmix_gpaw,get_nmix_qe,get_nmix_vasp,get_diag_gpaw,get_diag_qe,get_diag_vasp
          ,get_spinpol_gpaw,get_spinpol_qe,get_spinpol_vasp,get_dipole_gpaw
          ,get_dipole_qe,get_dipole_vasp,get_maxstep_gpaw,get_maxstep_qe,get_maxstep_vasp
          ,get_kpts_gpaw,get_kpts_qe,get_kpts_vasp]

################################################################
################################################################
################################################################

################################################################
################################################################
################################################################
struct_composition = Table('struct_composition'
        ,desc = 'Chemical Composition of a structure'
        ,cols = [Col('struct_id',             pk=True)
                ,Col('element_id',            pk=True)
                ,Col('has_fixed',             nn=True)
                ,Col('count',                 nn=True)
                ,Col('count_fixed',           nn=True)
                ,Col('count_norm',            nn=True)
                ,Col('frac',        Decimal(),nn=True)
                ,Col('frac_fixed',  Decimal(),nn=True)
                ,Col('fixed_norm',            nn=True)
                ]
        ,fks = [FK('struct_id', 'struct')
               ,FK('element_id','element')])

################################################################
surface = Table('surface'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id', pk=True)
                ,Col('vacuum',     Decimal())
                ,Col('pure_struct_id') # underlying structure
                ,Col('facet_h')
                ,Col('facet_k')
                ,Col('facet_l')
                ,Col('pure_struct_id_catalog')
                ,Col('facet_h_catalog')
                ,Col('facet_k_catalog')
                ,Col('facet_l_catalog')
                ]
        ,fks=[FK('struct_id','struct')
             ,FK('pure_struct_id','pure_struct')
             ,FK('pure_struct_id_catalog','pure_struct','pure_struct_id')])

bulk = Table('bulk'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id',    pk=True)
                ,Col('pure_struct_id_catalog')]
        ,fks=[FK('struct_id','struct')])

molecule = Table('molecule'
        ,desc = 'Subclass of struct'
        ,cols = [Col('struct_id',        pk=True)
                ,Col('pointgroup',  Varchar())]
        ,fks=[FK('struct_id','struct')])

########################################################################

################################################################################


similar_struct= Table('similar_struct'
    ,desc = "Store a precomputation for identifying related jobs"
    ,cols = [Col('struct_id',pk=True)
            ,Col('struct_id2',pk=True)]
    ,fks  = [FK('struct_id','struct')
            ,FK('struct_id2','struct','struct_id')])

############################################################################

calc_other = Table('calc_other'
        ,desc = """Less important DFT Calculator Parameters"""
        ,cols = [Col('calc_other_id',pk=True,auto=True)
                ,Col("kx")
                ,Col("ky")
                ,Col("kz")
                ,Col("fmax",            Decimal())
                ,Col("econv",           Decimal(10,7))
                ,Col("dw")
                ,Col("sigma",           Decimal(10,7))
                ,Col("nbands")
                ,Col("mixing",          Decimal())
                ,Col("nmix")
                ,Col("xtol",            Decimal(10,7))
                ,Col("strain",          Decimal())
                ,Col("gga",             Varchar())
                ,Col("luse_vdw")
                ,Col("zab_vdw",         Decimal())
                ,Col("nelmdl")
                ,Col("gamma")
                ,Col("dipol",           Varchar())
                ,Col("algo",            Varchar())
                ,Col("ibrion")
                ,Col("prec",            Varchar())
                ,Col("ionic_steps")
                ,Col("lreal",           Varchar())
                ,Col("lvhar")
                ,Col("diag",            Varchar())
                ,Col("spinpol")
                ,Col("dipole")
                ,Col("maxstep")
                ,Col("delta",           Decimal())
                ,Col("mixingtype",      Varchar())
                ,Col("bonded_inds",     Varchar())
                ,Col("energy_cut_off",  Decimal())
                ,Col("step_size",       Decimal())
                ,Col("spring",          Decimal())
                ,Col('cell_dofree',     Varchar())
                ,Col('cell_factor',     Decimal())
                ,Col("kpts",            Varchar())])


#############################################################################
vib_job = Table('vib_job'
        ,desc = 'Jobs that compute vibrational modes for a struct'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('struct_id','struct')])

dos_job = Table('dos_job'
        ,desc = 'Jobs that compute density of states for a struct'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('struct_id','struct')])

neb_job = Table('neb_job'
        ,desc = 'Jobs that compute NEB between two structures'
        ,cols = [Col('job_id',pk=True)
                ,Col('calc_id')
                ,Col('init_struct_id')
                ,Col('final_struct_id')]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',    'calc')
                ,FK('init_struct_id','struct','struct_id')
                ,FK('final_struct_id','struct','struct_id')])

bulkmod_job = Table('bulkmod_job'
        ,desc = 'Jobs that compute bulk modulus for a struct'
        ,cols = [Col('job_id',        pk=True)
                ,Col('calc_id')
                ,Col('struct_id')
                ,Col('bulkmod',    Decimal())]
        ,fks  = [FK('job_id',    'alljob')
                ,FK('calc_id',  'calc')
                ,FK('struct_id','struct')])

tables = [struct_composition
         ,surface
         ,bulk
         ,molecule
         ,calc_other
         ,similar_struct
         ,dos_job
         ,vib_job
         ,neb_job
         ,bulkmod_job]
##############################################################################
##############################################################################
##############################################################################


geograph = Rule('geograph'
    ,query    = """SELECT S.struct_id,S.raw from struct AS S WHERE S.geo_graph IS NULL"""
    ,template = Temp(tempfuncs = [TempFunc(geometric
                                         ,args=[Arg('json_to_traj')])
                                ,TempFunc(json_to_traj,args=[Arg('raw')])
                                ,TempFunc(graph_to_json
                                         ,'geo_graph'
                                         ,args=[Arg('mk_graph')])
                                ,TempFunc(mk_graph
                                         ,args = noIndex(['json_to_traj'
                                                           ,'include_frac'
                                                         ,'group_cut'
                                                         ,'min_bo'
                                                         ,'geometric']))]
                    ,constants = {'include_frac': Const(0.8)
                                 ,'group_cut'     : Const(0.3)
                                 ,'min_bo'         : Const(0.03)})
    ,metatemp = SimpleUpdate(struct,['geo_graph']))


elemental = Rule('elemental'
    ,metatemp=PureSQL(["""UPDATE struct AS S SET S.elemental =
                             S.n_atoms=1 AND S.system_type = 'molecule'
                             WHERE S.struct_id >0"""]))
fmax = Rule('fmax'
    ,metatemp=PureSQL(["""UPDATE traj AS T SET T.fmax =

                (SELECT
                    MAX(SQRT(TA.fx * TA.fx + TA.fy * TA.fy + TA.fz * TA.fz)) AS fmax
                FROM
                    traj AS T2
                JOIN
                    trajatom AS TA USING (job_id , traj_id , struct_id)
                JOIN
                    atom AS A USING (struct_id , atom_id)
                WHERE
                    A.constrained = 0
                GROUP BY job_id,traj_id)


                WHERE
                T2.job_id=T.job_id AND T2.traj_id=T.traj_id
                AND T.job_id >0 AND T.struct_id > 0"""]))

voronoi = Rule('voronoi'
    ,query      = """SELECT DISTINCT S.struct_id,S.raw
                     FROM struct AS S JOIN atom AS A USING (struct_id)
                     WHERE A.v3 IS NULL"""
    ,template     = Temp(tempfuncs=[TempFunc(get_voronoi
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(json_to_traj
                                          ,args = [Arg('raw')])])
    ,metatemp   = MetaTemp([Block(name='update_voronoi'
                                ,text="""UPDATE atom SET v3=%s,v4=%s,v5=%s,v6=%s
                                                        ,v7=%s,v8=%s,v9=%s,v10=%s
                                        WHERE struct_id = %s AND atom_id = %s"""
                                ,args = [Arg('get_voronoi',i) for i in range(1,9)]
                                        +[Arg('struct_id'),Arg('get_voronoi',0)])]))

populate_subjobs = Rule(name = 'subjobs'
        ,metatemp = PureSQL([
        """ INSERT INTO bulkmod_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'bulkmod'
                    AND J.job_id NOT IN (SELECT B.job_id from bulkmod_job as B)
                    AND NOT J.deleted"""
        ,""" INSERT INTO chargemol_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'chargemol'
                    AND J.job_id NOT IN (SELECT C.job_id from chargemol_job as C)
                    AND NOT J.deleted"""
        ,""" INSERT INTO vib_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'vib'
                    AND J.job_id NOT IN (SELECT V.job_id from vib_job as V)
                    AND NOT J.deleted"""
        ,""" INSERT INTO neb_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'neb'
                    AND J.job_id NOT IN (SELECT N.job_id from neb_job as N)
                    AND NOT J.deleted"""
        ,""" INSERT INTO dos_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type = 'dos'
                    AND J.job_id NOT IN (SELECT D.job_id from dos_job as D)
                    AND NOT J.deleted"""]))

populate_substructs = Rule(name = 'substructs'
    ,metatemp = PureSQL([
    """INSERT INTO molecule (struct_id)
        SELECT S.struct_id from struct AS S
        WHERE S.system_type = 'molecule'
        AND S.struct_id NOT IN (SELECT molecule.struct_id from molecule)"""
    ,"""INSERT INTO bulk (struct_id)
        SELECT S.struct_id from struct AS S
        WHERE S.system_type = 'bulk'
        AND S.struct_id NOT IN (SELECT bulk.struct_id from bulk)"""
    ,"""INSERT INTO surface (struct_id)
        SELECT S.struct_id from struct AS S
                        WHERE S.system_type = 'surface'
                        AND S.struct_id NOT IN (SELECT surface.struct_id from surface)"""]))


comp_cols = ['element_id','has_fixed'
            ,'count','count_fixed','count_norm'
            ,'frac','frac_fixed','fixed_norm']

# if any rows with a particular job have been added, we assume all have been added
populate_struct_composition = Rule(name='composition'
    ,query = """SELECT S.struct_id,S.raw
                FROM  struct AS S
                WHERE S.struct_id NOT IN (SELECT C.struct_id FROM struct_composition AS C)"""
    ,template    = SimpleTemp(load_comp,['raw'],comp_cols)
    ,metatemp    = SimpleInsert('struct_composition',['struct_id']+comp_cols)
    )

comp_strs = Rule(name='composition_strs'
    ,metatemp = PureSQL(["""
        UPDATE struct AS S SET S.composition_norm =
        (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count_norm,']') ORDER BY C.element_id ASC)
                FROM struct_composition as C
                WHERE C.struct_id = S.struct_id)

            ,S.composition=
              (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                    FROM struct_composition as C
                    WHERE C.struct_id = S.struct_id)

            ,S.metal_comp=
                (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count,']') ORDER BY C.element_id ASC)
                      FROM struct_composition as C
                      WHERE C.struct_id = S.struct_id
                      AND C.element_id NOT IN (1,2,6,7,8,9,10,16,17,18))
        WHERE S.struct_id > 0"""
            ]))


update_pointgroup = Rule(name='update_pointgroup'
    ,query = """SELECT M.struct_id,S.raw
                FROM molecule AS M
                    JOIN struct AS S USING (struct_id)
                WHERE M.pointgroup IS NULL"""
    ,template   = Temp(tempfuncs=[TempFunc(get_pointgroup
                                          ,name = 'pointgroup'
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(json_to_traj
                                          ,args = [Arg('raw')])])

    ,metatemp   = SimpleUpdate(tab  = molecule
                              ,cols = ['pointgroup']))

update_spacegroup = Rule(name='update_spacegroup'
    ,query = """SELECT S.struct_id,S.raw
                    FROM struct AS S
                WHERE S.spacegroup IS NULL"""
    ,template   = Temp(tempfuncs=[TempFunc(get_spacegroup
                                          ,name = 'spacegroup'
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(json_to_traj
                                          ,args = [Arg('raw')])])

    ,metatemp   = SimpleUpdate(tab  = struct
                              ,cols = ['spacegroup']))

#######################

# Calc other cols only relevant for VASP jobs
calc_other_pairs = {'prec'      : 'str'
                   ,'luse_vdw'  : 'bool'
                   ,'gga'       : 'bool'
                   ,'lvhar'     : 'bool'
                   ,'zab_vdw'   : 'float'
                   ,'algo'      : 'str'
                   ,'nelmdl'    : 'int'
                   ,'ibrion'    : 'int'
                   ,'lreal'     : 'str'
                   }

# Calc other cols only relevant for VASP
vasp_key_tempfuncs = [TempFunc(get_vasp_key
         ,name = x
         ,args = noIndex(['dftcode','log','key '+x,'cast '+y]))
                            for x,y in calc_other_pairs.items()]

# Calc other cols relevant for all three codes
dftdep_tempfuncs = [TempFunc(dftcode_dependent
                            ,name = x
                            ,args = noIndex(['dftcode'
                                            ,'log'
                                            ,'get_%s_gpaw'%x
                                            ,'get_%s_qe'%x
                                            ,'get_%s_vasp'%x
                                            ,'identity'
                                            ,'identity'])) for x in other_cols if x not in ['nbands','maxstep','kpts','gamma']]
# Calc other cols relevant for all three codes that depend on files other than log
dftdep_dict_tempfuncs = [TempFunc(dftcode_dependent
                                 ,name = x
                                 ,args = noIndex(['dftcode'
                                                  ,'files'
                                                  ,'get_%s_gpaw'%x
                                                  ,'get_%s_qe'%x
                                                  ,'get_%s_vasp'%x
                                                  ,'identity'
                                                  ,'identity'])) for x in  ['maxstep','kpts','gamma']]
# Calc other cols that can 'only' be scraped from params.json
get_pairs = {'fmax'              :'float'
            ,'strain'            :'float'
            ,'energy_cut_off'    :'float'
            ,'bonded_inds'       :'str'
            ,'step_size'         :'float'
            ,'spring'            :'float'
            ,'delta'             :'float'
            }

get_tempfuncs = [TempFunc(get
                         ,name = x
                         ,args = noIndex(['paramdict','key '+x,'cast '+y])) for x,y in get_pairs.items()]

qe_tempfuncs = [TempFunc(x,x.__name__[4:],args = noIndex(['pwinp','dftcode']))
                    for x in [get_cell_factor,get_cell_dofree]]

# All calc_other column names
calc_other_cols = (list(calc_other_pairs.keys())
                 + list (get_pairs.keys())
                 + [o for o in other_cols if o != 'kpts']
                 + ['kx','ky','kz','cell_dofree','cell_factor'])

calc_other_args = noIndex(calc_other_cols)
double_co_args  = [val for val in calc_other_args for _ in (0, 1)]



# Main rule
populate_calc_other = Rule(name='calc_other'
    ,query    = """SELECT R.job_id,J.log,J.pwinp,J.kptcar,J.paramdict,C.dftcode
                     FROM relax_job AS R
                     JOIN alljob AS J USING (job_id)
                     JOIN calc AS C  USING (calc_id)
                     WHERE R.calc_other_id IS NULL"""
    ,template = Temp(constants = merge_dicts([{'key ' +x:Const(x) for x in calc_other_pairs.keys()} # type: ignore
                                             ,{'cast '+y:Const(y) for y in calc_other_pairs.values()}
                                             ,{'key ' +x:Const(x) for x in get_pairs.keys()}
                                             ,{'cast '+y:Const(y) for y in get_pairs.values()}
                                             ,{'identity' : Const(identity)}
                                             ]+[{x.__name__ : Const(x)} for x in get_fns])

                    ,tempfuncs = [TempFunc(identity,'kx',args=[Arg('kpts',0)])
                                 ,TempFunc(identity,'ky',args=[Arg('kpts',1)])
                                 ,TempFunc(identity,'kz',args=[Arg('kpts',2)])
                                 ,TempFunc(get_dw,'dw',args=noIndex(['dftcode','log']))
                                 ,TempFunc(get_mixtype,'mixtype',args=noIndex(['dftcode','log']))
                                 ,TempFunc(pair,args=noIndex(['log','paramdict']))
                                 ,TempFunc(triple,'files',args=noIndex(['log','pwinp','kptcar']))
                                 ,TempFunc(dftcode_dependent
                                           ,name = 'nbands'
                                          ,args = noIndex(['dftcode'
                                                          ,'pair' # THIS IS WHAT IS DIFFERENT
                                                          ,'get_nbands_gpaw'
                                                          ,'get_nbands_qe'
                                                          ,'get_nbands_vasp'
                                                          ,'identity'
                                                          ,'identity']))]
                                 + dftdep_tempfuncs
                                 + dftdep_dict_tempfuncs
                                 + vasp_key_tempfuncs
                                 + get_tempfuncs
                                 + qe_tempfuncs
                    )
    ,metatemp = MetaTemp([Block(name = 'insert_calc'
                               ,text = """INSERT  into calc_other ({0})
                                         SELECT {1} FROM DUAL
                                        WHERE NOT EXISTS (SELECT (1) FROM calc_other WHERE {2})
                                        """.format(','.join(calc_other_cols)
                                                  ,','.join(['%s']*len(calc_other_cols))
                                                  ,' AND '.join(['({0}=%s OR ({0} IS NULL AND %s IS NULL ))'.format(x) for x in calc_other_cols]))
                                ,args = calc_other_args+double_co_args)
                        ,Block(name = 'get_calc'
                              ,text = """SELECT calc_other_id FROM calc_other WHERE {0}
                                        """.format(' AND '.join(['({0}=%s OR ({0} IS NULL AND %s IS NULL ))'.format(x) for x in calc_other_cols]))
                              ,args = double_co_args)
                        , Block(name = 'update'
                               ,text = "UPDATE relax_job SET calc_other_id = %s WHERE job_id=%s"
                               ,args = [Arg('get_calc',0),Arg('job_id')])]))

kden_cols = ['kptden_x','kptden_y','kptden_z']

kptden = Rule(name='kptden'
    ,query = """SELECT T.job_id,T.traj_id,O.kx,O.ky,O.kz%s
                FROM traj AS T
                    JOIN struct    AS S USING (struct_id)
                    JOIN cell        AS C USING (cell_id)
                    JOIN relax_job AS J USING (job_id)
                    JOIN calc_other AS O USING (calc_other_id)
                WHERE T.kptden_x IS null
            """%(''.join([',C.%s'%c for c in cell_cols]))
    ,template = SimpleTemp(get_kptden
                          ,['kx','ky','kz']+cell_cols
                          ,kden_cols)
    ,metatemp=SimpleUpdate(traj,kden_cols))


populate_struct_maps = Rule(name='struct_maps'
    ,metatemp=PureSQL(["""
    INSERT INTO chargemol_map (job_id,charge_id)

    SELECT
        F.job_id,CJ.job_id
    FROM
        chargemol_job AS CJ
            JOIN struct AS CS USING (struct_id)
            JOIN cell AS CC USING (cell_id)
            JOIN similar_struct AS SS USING (struct_id) -- pairs of structures with same sym, chemical symbols, and constrained atoms (all non-float data)
            JOIN finaltraj AS F ON F.struct_id = SS.struct_id2
            JOIN struct    AS S ON F.struct_id = S.struct_id
            JOIN cell      AS C ON C.cell_id = S.cell_id
    WHERE
        ABS(CC.a0 - C.a0) < 0.01
            AND ABS(CC.a1 - C.a1) < 0.01
            AND ABS(CC.a2 - C.a2) < 0.01
            AND ABS(CC.b0 - C.b0) < 0.01
            AND ABS(CC.b1 - C.b1) < 0.01
            AND ABS(CC.b2 - C.b2) < 0.01
            AND ABS(CC.c0 - C.c0) < 0.01
            AND ABS(CC.c1 - C.c1) < 0.01
            AND ABS(CC.c2 - C.c2) < 0.01
            AND NOT EXISTS( SELECT (1)
                FROM  atom AS A JOIN atom AS CA
                WHERE   A.struct_id = S.struct_id
                    AND CA.struct_id = CS.struct_id
                    AND A.atom_id = CA.atom_id
                    AND (  ABS(A.x - CA.x) > 0.01
                        OR ABS(A.y - CA.y) > 0.01
                        OR ABS(A.z - CA.z) > 0.01))

        ON DUPLICATE KEY UPDATE job_id=F.job_id"""])) # could replace 'chargemol' with 'vib' and reuse code


struct_strs = Rule('struct_strs'
    ,metatemp=PureSQL(["""
         UPDATE struct AS S SET S.str_symbols =
            (SELECT GROUP_CONCAT(A.element_id)
                FROM atom as A WHERE A.struct_id = S.struct_id)
  ,S.str_constraints =
    (SELECT GROUP_CONCAT(A.constrained)
        FROM atom as A WHERE A.struct_id = S.struct_id)

,S.symmetry = COALESCE((SELECT S.spacegroup
                FROM bulk AS B
                WHERE B.struct_id=S.struct_id)
            ,(SELECT M.pointgroup
                FROM molecule AS M
                WHERE M.struct_id=S.struct_id)
            ,'1')

    WHERE S.struct_id > 0            """]))

pop_similar_struct = Rule('pop_similar_struct'
    ,metatemp = PureSQL(["""INSERT INTO similar_struct  (struct_id,struct_id2)
                        SELECT S.struct_id,T.struct_id
                        FROM struct AS S JOIN struct AS T
                        WHERE S.str_symbols = T.str_symbols
                        AND S.str_constraints=T.str_constraints
                        AND S.symmetry = T.symmetry
                        ON DUPLICATE KEY UPDATE similar_struct.struct_id=similar_struct.struct_id"""]))

update_vacuum = Rule('update_vacuum'
    ,query = """SELECT U.struct_id,S.raw
                    FROM struct AS S JOIN surface AS U USING (struct_id)
                WHERE U.vacuum IS NULL"""
    ,template   = Temp(tempfuncs=[TempFunc(get_vacuum
                                          ,name = 'vacuum'
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(json_to_traj
                                          ,args = [Arg('raw')])])
    ,metatemp   = SimpleUpdate(tab  = surface
                              ,cols = ['vacuum']))

update_layer = Rule('update_layer'
    ,query = """SELECT U.struct_id,S.geo_graph
                FROM struct AS S
                JOIN surface AS U USING (struct_id)
                JOIN atom AS A USING (struct_id)
                GROUP BY U.struct_id
            HAVING SUM(A.layer is null)!=0"""
    ,template=Temp(tempfuncs=[TempFunc(json_to_graph
                                      ,args=[Arg('geo_graph')])
                            ,TempFunc(layers
                                    ,args=[Arg('json_to_graph')])
                            ,TempFunc(identity
                                    ,name='layer'
                                    ,args=[Arg('layers',1)])
                            ,TempFunc(identity
                                    ,name='atom_id'
                                    ,args=[Arg('layers',0)])])
    ,metatemp=SimpleUpdate(atom,['layer']))


pop_kpts = Rule('pop_kpts'
    ,metatemp=PureSQL(["""UPDATE calc_other AS C
                        SET  kpts = CONCAT('[', C.kx, ',', C.ky, ',', C.kz, ']')
                        WHERE C.calc_other_id > 0"""]))
rules = [geograph
             ,elemental
             ,voronoi
             ,populate_subjobs
             ,populate_substructs
             ,populate_struct_composition
             ,comp_strs
             ,update_pointgroup
             ,update_spacegroup
             ,update_vacuum
             ,populate_calc_other
             ,kptden
             ,populate_struct_maps
             ,struct_strs
             ,pop_similar_struct
             ,update_layer
             ,pop_kpts]



# Internal modules
from dbgen.support.datatypes.model      import Model
from dbgen.support.datatypes.func       import Func
from dbgen.new_inputs.objects           import uni
from dbgen.new_inputs.relations         import relations

###############################################################################
if __name__=='__main__':
    from dbgen.main import parser
    args = parser.parse_args()
    m = Model(uni,relations)
    m.run(reset = args.reset,catalog=args.catalog)



from typing import Optional, Tuple
from re import finditer, MULTILINE, search, DOTALL


def parse_line(string: str, substr: str, index: int = 0
               ) -> Optional[str]:
    """
    Returns the n'th line containing substring
    Any negative index will return last one.
    """
    iter = finditer(substr + ".*$", string, MULTILINE)
    found = False

    for match in iter:
        if index == 0:
            return match[0]
        else:
            index -= 1
            found = True
    if found:
        return match[0]  # negative input for index
    else:
        return None


def btw(s: str, begin: str, end: str, off: int = 0
        ) -> Tuple[str, int]:
    result = search('%s(.*?)%s' % (begin, end), s[off:], DOTALL)
    if result:
        if result.group(1) is None:
            import pdb
            pdb.set_trace()
        return result.group(1), result.end() + off
    else:
        return '', 0




# External Modules
import os
# Internal Modules
from dbgen.support.utils              import sqlite_execute,mkInsCmd
from dbgen.support.datatypes.table	import Table,Col,FK
from dbgen.support.datatypes.sqltypes	import Varchar

"""
Define the schema for a SQLite meta.db, which can be used to represent the
entire DbGen process.
"""

################################################################################
tables = [Table('tab'
            ,cols=[Col('tab_id',        pk = True)
                  ,Col('name',Varchar(),nn = True)
                  ,Col('desc',Varchar(),nn = True)])
    ,Table('views'
            ,cols=[Col('view_id',           pk = True)
                  ,Col('name',  Varchar(),  nn = True)
                  ,Col('query', Varchar(),  nn = True)])
    ,Table('col'
        ,cols=[Col('tab_id',        pk = True)
              ,Col('col_id',        pk = True)
              ,Col('name',Varchar(),nn = True)
              ,Col('type',Varchar(),nn = True)
              ,Col('nnull',         nn = True)
              ,Col('pk',            nn = True)
              ,Col('uniq',          nn = True)
              ,Col('auto',          nn = True)
              ,Col('ind',           nn = True)
              ,Col('virt',  Varchar())
              ,Col('dfault',Varchar())
              ]
        ,fks=[FK('tab_id','tab')])

    ,Table('fk'
        ,cols=[Col('tab_id', pk = True)
              ,Col('fk_id',  pk = True)]
        ,fks=[FK('tab_id','tab')])

    ,Table('fk_cols'
        ,cols=[Col('tab_id',  pk = True)
              ,Col('fk_id',   pk = True)
              ,Col('col_id',  pk = True)
              ,Col('to_tab',  nn = True)
              ,Col('to_col',  nn = True)]
        ,fks=[FK(['tab_id','fk_id'],'fk')
             ,FK(['tab_id','col_id'],'col')
             ,FK(['to_tab','to_col'],'col',['tab_id','col_id'])])

    ,Table('func'
        ,cols=[Col('func_id',               pk = True)
              ,Col('name',        Varchar(),nn = True)
              ,Col('pth',         Varchar(),nn = True)
              ,Col('source',      Varchar(),nn = True)
              ,Col('docstring',   Varchar(),nn = True)
              ,Col('language',    Varchar(),nn = True)
              ,Col('inTypes_json',Varchar(),nn = True)
              ,Col('outType_json',Varchar(),nn = True)
              ,Col('inTypes',     Varchar(),nn = True)
              ,Col('outType',     Varchar(),nn = True)
              ,Col('n_in',                  nn = True)
              ,Col('n_out',                 nn = True)])

    ,Table('rule'
        ,cols=[Col('rule_id',      pk=True)
              ,Col('name',   Varchar(), nn=True)
              ,Col('query',  Varchar())])

    ,Table('tempfunc'
        ,cols=[Col('rule_id',              pk = True)
              ,Col('tempfunc_id',               pk = True)
              ,Col('func_id',                   nn = True)
              ,Col('name',          Varchar(),  nn = True)]
        ,fks = [FK('rule_id','rule')
               ,FK('func_id','func')])

    ,Table('tempfuncarg'
        ,cols=[Col('rule_id',              pk = True)
              ,Col('tempfunc_id',               pk = True)
              ,Col('tempfuncarg_id',            pk = True)
              ,Col('name',          Varchar(),  nn = True)
              ,Col('ind')]
        ,fks=[FK(['rule_id','tempfunc_id'],'tempfunc')])

    ,Table('constants'
        ,cols=[Col('rule_id',          pk = True)
              ,Col('const_id',              pk = True)
              ,Col('name',      Varchar(),  nn = True)
              ,Col('datatype',  Varchar(),  nn = True)
              ,Col('val',       Varchar(),  nn = True)]
        ,fks = [FK('rule_id','rule')])

    ,Table('libfuncs'
        ,cols=[Col('rule_id',              pk = True)
              ,Col('const_id',                  pk = True)
              ,Col('name',          Varchar(),  nn = True)
              ,Col('func_id',                   nn = True)]
        ,fks = [FK('func_id','func')
               ,FK(['rule_id'],'rule')])

    ,Table('blockarg'
        ,cols=[Col('rule_id',              pk = True)
              ,Col('block_id',                  pk = True)
              ,Col('blockarg_id',               pk = True)
              ,Col('name',          Varchar(),  nn = True)
              ,Col('ind')]
        ,fks=[FK(['rule_id','block_id'] ,'block')])

    ,Table('block'
        ,cols = [Col('rule_id',            pk = True)
                ,Col('block_id',                pk = True)
                ,Col('text',        Varchar(),  nn = True)
                ,Col('name',        Varchar(),  nn = True)]
        ,fks = [FK('rule_id','rule')])

    ,Table('namespace'
        ,cols=[Col('namespace_id',              pk = True)
              ,Col('name',          Varchar(),  nn = True,uniq = True)
              ,Col('type',          Varchar(),  nn = True,uniq = True)])

    ,Table('config'
        ,cols=[Col('config_id',             pk = True)
              ,Col('verbose',               nn = True)
              ,Col('retry',                 nn = True)
              ,Col('simulate',              nn = True)
              ,Col('run_only',  Varchar())
              ,Col('run_except',Varchar())]
        ,fks=[FK('connection_id','connection')])

    ,Table('connection'
        ,cols=[Col('connection_id' ,        pk = True)
              ,Col('name',      Varchar())
              ,Col('hostname',  Varchar(),  nn = True,uniq = True)
              ,Col('user',      Varchar(),  nn = True,uniq = True)
              ,Col('port',                  nn = True,uniq = True)
              ,Col('database',  Varchar(),  nn = True,uniq = True)
              ,Col('password',  Varchar(),  nn = True,uniq = True)])
        ]

###############################################################################
# Main
#-----
def create_empty(pth   : str
                ,local : bool = False
                ) -> None:
    """
    Creates a brand new database file containing no data (except default config)
    """
    if os.path.exists(pth):
        print('removing db at ',pth)
        os.remove(pth)

    # Create all tables
    #----------------
    for t in tables:
        sqlite_execute(pth,t.create_table(sqlite=True))

    # Populate two possible MySQL connections: localhost vs SUNCAT
    #-----------------------------------------------------------------
    insertConn = mkInsCmd('connection',['name','hostname','port'
                                       ,'user','password','database']
                        ,sqlite=True)
    user = os.environ['USER']
    sqlite_execute(pth,insertConn,['localhost',"127.0.0.1",3306,user,user, "suncat"])

    # Initialize Config
    #----------------
    config = 1 if local else 2
    init_config = mkInsCmd('config'
                          ,['verbose','retry','simulate','connection_id']
                          ,sqlite = True)

    sqlite_execute(pth,init_config,[1,1,0,config])


if __name__=='__main__':
    create_empty('temp.db')



# External Modules
from copy import deepcopy
# Internal modules
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.rule import Rule
from dbgen.support.datatypes.metatemp  import PureSQL
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal
from dbgen.core.lists                  import flatten


##########################################################################################

#Refeng tables
#-------------
ref_scheme = Table('ref_scheme'
    ,desc = 'A set of rules to determine when a struct is a valid reference state.'
    ,cols =[Col('ref_scheme_id', pk = True,auto = True)
           ,Col('name',Varchar(),    nn = True,uniq = True)])

ref_scheme_elem = Table('ref_scheme_elem'
    ,desc = 'For each reference scheme and each element, specify the symmetry needed'
    ,cols = [Col('ref_scheme_id',    pk=True)
            ,Col('element_id',        pk=True)
            ,Col('symmetry',Varchar(),    nn=True)]
    ,fks  = [FK('ref_scheme_id','ref_scheme')
            ,FK('element_id',   'element')])

ref_scheme_stoich = Table('ref_scheme_stoich'
    ,desc = """For each reference scheme and each element, specify the reduced stoichiometry needed.
               This is essentially a constraint on which jobs are regarded as reference states for a given element."""
    ,cols=[Col('ref_scheme_id',        pk=True)
          ,Col('element_id',        pk=True)
          ,Col('component_id',        pk=True)
          ,Col('count',                nn=True)]
    ,fks = [FK(['ref_scheme_id','element_id'],'ref_scheme_elem')
           ,FK('component_id',  'element',    'element_id')])

ref_scheme_reqs =Table('ref_scheme_reqs'
    ,desc = """For each reference scheme and each element, specify the combination of reference jobs required to calculate the energy per atom.
This is essentially 'precomputing' the matrix problem that would normally be need to be solved.
If element 8 has (8,1),(1,-2) it would mean:
    The energy per atom for an oxygen atom is 1* normalized energy of job identified as reference for oxygen
                                             -2* normalized energy of job identified as reference for hydrogen
Where the 'normalized energy' of a reference job for element X is the raw energy of that job divided by its count for X
(This normalization trick is to allow us to use primitive/unit cells interchangably)"""
    ,cols=[Col('ref_scheme_id',        pk=True)
          ,Col('element_id',        pk=True)
          ,Col('component_id',        pk=True)
          ,Col('coef',Decimal(),        nn=True)]
    ,fks = [FK(['ref_scheme_id','element_id'],'ref_scheme_elem')
           ,FK('component_id',   'element',   'element_id')])

ref_scheme_refstruct = Table('ref_scheme_refstruct'
    ,desc = 'Warehouse of every structure that is a reference for some refscheme'
    ,cols = [Col('ref_scheme_id',    pk=True)
            ,Col('element_id',       pk=True)
            ,Col('struct_id',        pk=True)]
    ,fks  = [FK(['ref_scheme_id','element_id'],'ref_scheme_elem')
            ,FK('struct_id','struct')])

ref_scheme_job = Table('ref_scheme_job'
    ,desc= """For each element x calc x component, list the job and other useful
                info for calculating energies """
    ,cols = [Col('ref_scheme_id',    pk=True)  #  are
            ,Col('element_id',       pk=True)  #  absolutely
            ,Col('component_id',     pk=True)  #  necessary
            ,Col('job_id',           pk=True)  #  !
            ,Col('calc_id',          nn=True)
            ,Col('calc_other_id',    nn=True)
            ,Col('coef',             nn=True)
            ,Col('energy',Decimal(),   nn=True)
            ]
    ,fks = [FK(['ref_scheme_id','element_id','component_id'],'ref_scheme_reqs')
           ,FK('job_id','relax_job')
           ,FK('calc_id','calc')
           ,FK('calc_other_id','calc_other')])


tables = [ref_scheme,ref_scheme_elem, ref_scheme_job
         ,ref_scheme_refstruct,ref_scheme_reqs
         ,ref_scheme_stoich]


##########################################################################################
##########################################################################################
##########################################################################################
syms = {'2'   : [15]
       ,'14'  : [34]
       ,'64'  : [31,53]
       ,'70'  : [16]
       ,'139' : [49]
       ,'141' : [50]
       ,'152' : [52]
       ,'166' : [5,33,51,62,80]
       ,'194' : [4,12,21,22,27,30,39,40,43,44,48,57,58,59,60,61,64,65,66,67,68,69,71,72,75,76,81]
       ,'225' : [13,20,28,29,32,38,45,46,47,70,77,78,79,82]
       ,'227' : [14]
       ,'229' : [3,11,19,23,24,25,26,37,41,42,55,56,63,73,74]
       ,'C*v' : [6]
       ,'C2v' : [8]
       ,'D*h' : [1,7,9,17,35]
       ,'K*h' : [2,10,18,36,54]}

syms2 = deepcopy(syms)
syms2['227'].append(syms2['C*v'].pop())
syms2['D*h'].append(syms2['C2v'].pop())

allelems = flatten(list(syms.values()))
ref_scheme_elem_inserts = """INSERT INTO ref_scheme_elem (ref_scheme_id,element_id,symmetry)
                             VALUES %s ON DUPLICATE KEY UPDATE ref_scheme_id=ref_scheme_id
                             """%(','.join(flatten([["(1,%s,'%s')"%(x,y)
                                        for x in xs] for y,xs in syms.items()])
                                        +flatten([["(2,%s,'%s')"%(x,y)
                                                       for x in xs] for y,xs in syms2.items()])))
ref_scheme_stoich_inserts = """INSERT INTO ref_scheme_stoich (ref_scheme_id,element_id,component_id,count)
                               VALUES (1,6,6,1),(1,6,8,1),(1,8,1,2),(1,8,8,1)
                               ,%s ON DUPLICATE KEY UPDATE ref_scheme_id=ref_scheme_id
                               """%(','.join(['(1,{0},{0},1)'.format(i) for i in range(1,83) if i not in [6,8]]
                                               +['(2,{0},{0},1)'.format(i) for i in range(1,83)]))

ref_scheme_req_inserts = """INSERT INTO ref_scheme_reqs (ref_scheme_id,element_id,component_id,coef)
                               VALUES (1,6,6,1),(1,6,8,-1),(1,6,1,2)
                                        ,(1,8,1,-2),(1,8,8,1)
                               ,%s ON DUPLICATE KEY UPDATE ref_scheme_id=ref_scheme_id
                               """%(','.join(['(1,{0},{0},1)'.format(i) for i in range(1,83) if i not in [6,8]]
                                               +['(2,{0},{0},1)'.format(i) for i in range(1,83)]))

default_refscheme = Rule('default_refscheme'
    ,metatemp=PureSQL(["""INSERT INTO ref_scheme (name) VALUES ('default'),('pure')
                            ON DUPLICATE KEY UPDATE ref_scheme_id=ref_scheme_id"""
                      ,ref_scheme_elem_inserts
                      ,ref_scheme_stoich_inserts
                      ,ref_scheme_req_inserts]))


populate_refs = Rule('populate_refs'
     ,metatemp = PureSQL([
    """INSERT INTO ref_scheme_refstruct (ref_scheme_id
                                ,element_id
                                ,struct_id)
    SELECT R.ref_scheme_id
                        ,R.element_id
                        ,S.struct_id
    FROM ref_scheme_elem  AS R
      JOIN ref_scheme_stoich  AS RS ON RS.ref_scheme_id = R.ref_scheme_id AND  RS.element_id = R.element_id
      JOIN  struct   AS S ON R.symmetry=S.symmetry
      JOIN struct_composition AS C ON C.struct_id=S.struct_id AND C.element_id = RS.component_id

        WHERE S.struct_id in (SELECT struct_id from finaltraj)
        GROUP BY S.struct_id,R.ref_scheme_id,R.element_id
        HAVING GROUP_CONCAT(CONCAT('[',RS.component_id,',',RS.count,']'))
              = GROUP_CONCAT(CONCAT('[',C.element_id,',',C.count_norm,']'))

        ON DUPLICATE KEY UPDATE ref_scheme_id=ref_scheme_id"""]))

pop_rsjob = Rule('populate_refscheme_job'
    ,metatemp=PureSQL(["""
        INSERT INTO ref_scheme_job (ref_scheme_id
                                   ,element_id
                                   ,component_id
                                   ,coef
                                   ,job_id
                                   ,energy
                                   ,calc_id
                                   ,calc_other_id)
        SELECT R.ref_scheme_id
               ,R.element_id
               ,R.component_id
               ,R.coef
               ,F.job_id
               ,F.energy
               ,J.calc_id
               ,J.calc_other_id
        FROM
        ref_scheme_reqs           AS R
        JOIN ref_scheme_refstruct AS S ON R.component_id=S.element_id AND R.ref_scheme_id = S.ref_scheme_id
        JOIN finaltraj            AS F USING (struct_id) -- (job,step#,energy)
        JOIN relax_job            AS J USING (job_id)

        ON DUPLICATE KEY UPDATE ref_scheme_id=R.ref_scheme_id
        """]))



rules = [default_refscheme
             ,populate_refs
             ,pop_rsjob]



# External Modules
from os import environ
# Internal Modules
from dbgen.inputs.species              import nick_dict
from dbgen.core.lists                  import merge_dicts
from dbgen.support.datatypes.rule      import Rule
from dbgen.support.datatypes.temp      import Temp,TempFunc,Const
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.metatemp  import MetaTemp,Block,SimpleUpdate
from dbgen.support.utils               import mkInsCmd,mkUpdateCmd
from dbgen.inputs.basics               import alljob

from dbgen.scripts.Pure.Misc.get                import get
from dbgen.scripts.Pure.Misc.get_               import get_
from dbgen.scripts.Pure.Load.get_catalog_ads    import get_catalog_ads
from dbgen.scripts.Pure.Load.get_catalog_facet  import get_catalog_facet
"""
Anything that seems Catalog-specific (absolutely useless for jobs generated
elsewhere) should go here
"""
################################################################################
pypth = environ['DBGEN_ROOT']+'/scripts/'

functions  = {
    'get'                : pypth+"Pure/Misc/get.py"
    ,'get_'              : pypth+"Pure/Misc/get_.py"
    ,'get_catalog_facet' : pypth+"Pure/Load/get_catalog_facet.py"
    ,'get_catalog_ads'   : pypth+"Pure/Load/get_catalog_ads.py"
    }
################################################################################

# Mapping from informal structure names given by user to specific pure_struct's
struct_dict = merge_dicts([{v:k,v+'_unit':k,v+'-unit':k} for k,v in nick_dict.items()])
# Additional cases
struct_dict['2DMaterial'] = 'N/A'
struct_dict['hexagonal'] = struct_dict['hcp']

pop_bulk_struct_catalog = Rule('pop_bulk_struct_catalog'
    ,query = """SELECT B.struct_id,J.paramdict
                FROM bulk AS B
                JOIN struct AS S USING (struct_id)
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict != ''
                AND S.pure_struct_id_catalog IS NULL
                AND NOT J.deleted"""
    ,template = Temp(constants={'struct_dict'   : Const(struct_dict)
                               ,'structure_key' : Const('structure')
                               ,'str'           : Const('str')
                               ,'unknown'       : Const('unknown')}
                  ,tempfuncs=[TempFunc(get
                                      ,name='nickname'
                                      ,args=noIndex(['paramdict'
                                                    ,'structure_key'
                                                    ,'str']))
                             ,TempFunc(get_
                                      ,'pure_struct_name'
                                      ,args=noIndex(['struct_dict'
                                                    ,'nickname'
                                                    ,'unknown']))])
    ,metatemp = MetaTemp([Block('insert_pure_struct'
                               ,mkInsCmd('pure_struct',['name'])
                               ,args=[Arg('pure_struct_name')])
                         ,Block('query_pure_struct'
                             ,"""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                             ,args=[Arg('pure_struct_name')])
                        ,Block('update_fk'
                             ,mkUpdateCmd('struct',['pure_struct_id_catalog'],['struct_id'])
                             ,args=noIndex(['query_pure_struct','struct_id']))]))

pop_surf_struct_catalog = Rule('pop_surf_struct_catalog'
    ,query="""SELECT S.struct_id,J.paramdict
                FROM surface AS S
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict != ''
                AND S.facet_l_catalog IS NULL
                AND NOT J.deleted """
    ,template=Temp(constants={'struct_dict'     : Const(struct_dict)
                             ,'structure_key'   : Const('structure')
                             ,'str'             : Const('str')
                             ,'unknown'         : Const('unknown')}
                  ,tempfuncs=[TempFunc(get
                                      ,name='nickname'
                                      ,args=noIndex(['paramdict'
                                                    ,'structure_key'
                                                    ,'str']))

                             ,TempFunc(get_catalog_facet
                                     ,args=[Arg('paramdict')])

                             ,TempFunc(get_
                                      ,'pure_struct_name'
                                      ,args=noIndex(['struct_dict'
                                                    ,'nickname'
                                                    ,'unknown']))])
    ,metatemp = MetaTemp([Block('insert_pure_struct'
                             ,mkInsCmd('pure_struct',['name'])
                             ,args=[Arg('pure_struct_name')])
                       ,Block('query_pure_struct'
                             ,"""SELECT pure_struct_id FROM pure_struct WHERE name=%s"""
                             ,args=[Arg('pure_struct_name')])
                       ,Block('update_fk'
                             ,mkUpdateCmd('surface'
                                         ,['pure_struct_id_catalog'
                                          ,'facet_h_catalog'
                                          ,'facet_k_catalog'
                                          ,'facet_l_catalog']
                                         ,['struct_id'])
                             ,args=[Arg('query_pure_struct')
                                    ,Arg('get_catalog_facet',0)
                                    ,Arg('get_catalog_facet',1)
                                    ,Arg('get_catalog_facet',2)
                                    ,Arg('struct_id')])]))

pop_ads_catalog = Rule('pop_ads_catalog'
    ,query = """SELECT J.job_id,S.struct_id,J.paramdict
                FROM surface AS S
                JOIN traj USING (struct_id)
                JOIN alljob AS J USING (job_id)
                WHERE J.paramdict like '%%adsorbates%%'
                AND J.ads_catalog IS NULL
                AND NOT J.deleted"""
    ,template=Temp(constants={'ads': Const('adsorbates')
                             ,'str': Const('str')}
        ,tempfuncs=[TempFunc(get
                            ,name='ads_json'
                            ,args=noIndex(['paramdict'
                                            ,'ads'
                                            ,'str']))
                   ,TempFunc(get_catalog_ads
                            ,args=[Arg('ads_json')])])
    ,metatemp=MetaTemp([Block('insert'
                            ,mkUpdateCmd('atom',['struct_adsorbate_id_catalog'],['struct_id','atom_id'])
                            ,args=[Arg('get_catalog_ads',0)
                                  ,Arg('struct_id')
                                  ,Arg('get_catalog_ads',1)])
                    ,Block('insert2'
                        ,mkUpdateCmd('alljob',['ads_catalog'],['job_id'])
                        ,args=noIndex(['ads_json','job_id']))]))

pop_struct_catalog = Rule('pop_struct_catalog'
    ,query = """SELECT J.job_id,J.paramdict
                FROM alljob AS J
                WHERE J.paramdict like '%%structure%%'
                AND J.structure_catalog IS NULL
                AND NOT J.deleted """
    ,template=Temp(constants={'struct': Const('structure')
                             ,'str'   : Const('str')}
        ,tempfuncs=[TempFunc(get
                            ,name = 'structure_catalog'
                            ,args = noIndex(['paramdict'
                                            ,'struct'
                                            ,'str']))])
    ,metatemp = SimpleUpdate(alljob,['structure_catalog']))

rules = [pop_bulk_struct_catalog
             ,pop_surf_struct_catalog
             ,pop_ads_catalog
             ,pop_struct_catalog]




# Internal Modules

from dbgen.inputs.basics      import (tables as basic_tables
                                     ,rules as basic_rules
                                     ,views as basic_views)

from dbgen.inputs.adsorbates  import (tables as ads_tables
                                     ,rules  as ads_rules
                                     ,views  as ads_views)

from dbgen.inputs.experiment  import (tables     as expt_tables
                                           ,rules as expt_rules)

from dbgen.inputs.extra       import (tables     as x_tables
                                           ,rules as x_rules)

from dbgen.inputs.graphs      import (tables     as grph_tables
                                           ,rules as grph_rules)

from dbgen.inputs.learning    import (tables     as ml_tables
                                           ,rules as ml_rules)

from dbgen.inputs.references  import (tables     as ref_tables
                                           ,rules as ref_rules)

from dbgen.inputs.rxn         import (tables     as rxn_tables
                                           ,rules as rxn_rules)

from dbgen.inputs.species     import (tables     as sp_tables
                                           ,rules as sp_rules)

from dbgen.inputs.catalog     import (rules  as cat_rules)


from dbgen.support.datatypes.dbg import DBG
################################################################################

tables = (ads_tables + basic_tables + expt_tables + x_tables   + grph_tables
        + ml_tables  + grph_tables  + ml_tables   + ref_tables + rxn_tables
        + sp_tables)

rules = (ads_rules  + basic_rules + expt_rules + x_rules
            + grph_rules + ml_rules    + grph_rules + ml_rules
            + ref_rules  + rxn_rules   + sp_rules   + cat_rules)

views = basic_views + ads_views


dbg = DBG(tables = tables
         ,rules  = rules
         ,views  = views)



# External
from os import environ
# Internal
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.temp      import Temp,TempFunc,SimpleTemp
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.metatemp  import SimpleUpsert,SimpleUpdate,SimpleInsert,PureSQL,MetaTemp,Block
from dbgen.support.datatypes.rule      import Rule
from dbgen.support.datatypes.sqltypes  import Text,Decimal
from dbgen.support.utils               import addQs,mkInsCmd
from dbgen.core.lists                  import flatten

##########################################################################################
##########################################################################################
##########################################################################################
# CONSTANTS
###########
pypth = environ['DBGEN_ROOT']+'/scripts/'

cell_args = [Arg('cell',i) for i in range(9)]
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']
from dbgen.scripts.IO.parse_chargemol        import parse_chargemol
from dbgen.scripts.IO.anytraj                import anytraj
from dbgen.scripts.Pure.Atoms.traj_to_json   import traj_to_json
from dbgen.scripts.Pure.Atoms.json_to_traj   import json_to_traj
from dbgen.scripts.Pure.Atoms.get_cell       import get_cell
from dbgen.scripts.Pure.Atoms.get_atom       import get_atom
from dbgen.scripts.Pure.Graph.mk_graph       import mk_graph
from dbgen.scripts.Pure.Graph.graph_to_json  import graph_to_json
from dbgen.scripts.Pure.Misc.identity        import identity
##########################################################################################


bond = Table('bond'
    ,desc="A bond between two atoms identified by chargemol"
    ,cols = [Col('job_id',              pk = True)
            ,Col('struct_id',           pk = True)
            ,Col('atom1',               pk = True)
            ,Col('atom2',               pk = True)
            ,Col('bond_id',             pk = True)
            ,Col('bondorder',Decimal(), nn = True)
            ,Col('distance', Decimal(), nn = True)
            ,Col('pbc_x',               nn = True)
            ,Col('pbc_y',               nn = True)
            ,Col('pbc_z',               nn = True)
            ,Col('h_bond')]
    ,fks  = [FK('job_id',             'chargemol_job',  'job_id')
            ,FK(['struct_id','atom1'],'atom',           ['struct_id','atom_id'])
            ,FK(['struct_id','atom2'],'atom',           ['struct_id','atom_id'])])

chargemol_atom = Table('chargemol_atom'
    ,desc = "Properties of an atom + a chargemol job (e.g. DDEC charge)"
    ,cols = [Col('job_id',                  pk = True)
            ,Col('struct_id',               pk = True)
            ,Col('atom_id',                 pk = True)
            ,Col('charge',      Decimal(),  nn = True)]
    ,fks  = [FK('job_id',               'chargemol_job')
            ,FK(['struct_id','atom_id'],'atom')])

graph_params = Table('graph_params'
    ,desc = "Settings needed to create a graph out of a chargemol job"
    ,cols = [Col('graph_params_id',       pk = True,auto = True)
            ,Col('include_frac',Decimal(),nn = True,uniq = True)
            ,Col('group_cut',   Decimal(),nn = True,uniq = True)
            ,Col('min_bo',      Decimal(),nn = True,uniq = True)])


graph = Table('graph'
    ,desc = """Properties of graphs produced by chargemol analysis
            (combination of chargemol job + graph parameters)"""
    ,cols = [Col('job_id',            pk = True)
            ,Col('graph_params_id',   pk = True)
            ,Col('raw',Text(),        nn = True)
            ,Col('num_nodes')
            ,Col('num_edges')
            ,Col('num_hbonds')]
    ,fks  = [FK('job_id','chargemol_job')
            ,FK('graph_params_id','graph_params')])
#############################################################################
chargemol_map = Table('chargemol_map'
    ,desc = 'Mapping table to relate a job to plausible chargemol_jobs'
    ,cols = [Col('job_id',   pk = True)
            ,Col('charge_id',pk = True)]
    ,fks   = [FK('job_id',      'alljob')
             ,FK('charge_id','chargemol_job','job_id')])

chargemol_job = Table('chargemol_job'
        ,desc = 'Jobs that compute bond order analysis with Chargemol'
        ,cols = [Col('job_id',   pk=True)
                ,Col('calc_id')
                ,Col('struct_id')
                ,Col('bonddata',Text('long'))]
        ,fks  = [FK('job_id',   'alljob')
                ,FK('calc_id',  'calc')
                ,FK('struct_id','struct')])

tables = [bond,chargemol_atom,graph_params,graph,chargemol_map,chargemol_job]
##########################################################################################
##########################################################################################
##########################################################################################

get_chargemol_struct = Rule(name='get_chargemol_struct'
    ,query = """SELECT chargemol_job.job_id,J.stordir,J.anytraj
                FROM chargemol_job  JOIN alljob as J USING (job_id)
                WHERE chargemol_job.struct_id is null
                AND NOT J.deleted"""
    ,template = Temp(tempfuncs = [TempFunc(traj_to_json
                                           ,name='raw'
                                          ,args=[Arg('anytraj')])
                                  ,TempFunc(get_cell
                                                ,name = 'cell'
                                             ,args = [Arg('raw')])
                                 ,TempFunc(get_atom
                                           ,args = [Arg('raw')])]) # returns ([ax]m,[ay]m,...)

    ,metatemp = MetaTemp([Block(name = 'insert_cell'
                               ,text = mkInsCmd('cell',cell_cols,sqlite=False)
                               ,args = cell_args)

                         ,Block(name='query_cell'
                                ,text=('SELECT C.cell_id FROM cell AS C WHERE '
                                      +addQs(['C.%s'%x for x in cell_cols],' AND '))
                                ,args=cell_args)

                                 ,Block(name  = 'insert_struct'
                                        ,text = """INSERT INTO struct (raw,rawhash,cell_id)
                                                   VALUES (%s,SHA2(%s,512),%s)
                                                ON DUPLICATE KEY UPDATE cell_id=cell_id"""
                                       ,args = [Arg('raw'),Arg('raw'),Arg('query_cell',0)])

                         ,Block(name='query_struct'
                                 ,text='SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
                                ,args=[Arg('raw')])

                         ,Block(name = 'insert_atom'
                                ,text = mkInsCmd('atom',['struct_id','atom_id','element_id'
                                                      ,'x','y','z','constrained','magmom','tag']
                                                ,sqlite=False)
                               ,args= [Arg('query_struct')]
                                        +[Arg('get_atom',i) for i in range(1,9)])


                        ,Block(name='insert'
                             ,text="UPDATE chargemol_job SET struct_id=%s WHERE job_id=%s"
                             ,args=[Arg('query_struct'),Arg('job_id')])]))

parse_chgml_output = ['atom1','atom2','bond_id','pbc_x','pbc_y','pbc_z'
                     ,'distance','bondorder','atom_id','charge']

bond_cols = ['job_id','struct_id','atom1','atom2','bond_id'
            ,'pbc_x','pbc_y','pbc_z','distance','bondorder']

chgmol_atom_cols = ['job_id','struct_id','atom_id','charge']

unpackargs = [TempFunc(identity
                      ,name = x
                      ,args = [Arg('parse_chargemol',i)])
              for i,x in enumerate(parse_chgml_output)]

populate_bond = Rule('populate_bond'
    ,query = """SELECT C.job_id,S.struct_id,J.stordir
                FROM chargemol_job AS C
                    JOIN alljob        AS J USING (job_id)
                    JOIN struct AS S USING (struct_id)
                WHERE (C.job_id,S.struct_id) NOT IN (SELECT B.job_id,B.struct_id FROM bond AS B)
                AND NOT J.deleted"""
    ,template   = Temp(tempfuncs = unpackargs
                                 + [TempFunc(parse_chargemol
                                            ,args=noIndex(['stordir','anytraj']))
                                   ,TempFunc(anytraj
                                               ,args=[Arg('stordir')])] )
    ,metatemp   = MetaTemp([Block(name = 'insert_bond'
                                 ,text = mkInsCmd('bond',bond_cols,sqlite=False)
                                 ,args = noIndex(bond_cols))
                            ,Block(name   = 'insert_atom'
                                    ,text = mkInsCmd('chargemol_atom',chgmol_atom_cols,sqlite=False)
                                    ,args = noIndex(chgmol_atom_cols))]))

set_default_graph = Rule('default_graph_params'
    ,query = """INSERT INTO graph_params (include_frac,group_cut,min_bo)
                VALUES (0.8,0.3,0.03) ON DUPLICATE KEY UPDATE include_frac=include_frac""")

######
populate_bonddata = Rule('populate_bonddata'
    ,query="""SELECT job_id
                    ,CONCAT('[',GROUP_CONCAT('[',atom1,',',atom2
                                            ,',',bondorder
                                            ,',',pbc_x,',',pbc_y,',',pbc_z
                                            ,']'),']') as bonddata
                  FROM bond as B
            GROUP BY job_id"""
    ,metatemp=SimpleUpdate(chargemol_job,['bonddata']))
######
mk_graph_args = ['rawstruct','include_frac','group_cut','min_bo','bonddata']

populate_graph = Rule('populate_graph'
    ,query = """SELECT GP.graph_params_id
                      ,C.job_id
                      ,S.raw AS rawstruct
                      ,GP.include_frac
                      ,GP.group_cut
                      ,GP.min_bo
                      ,C.bonddata
                      FROM chargemol_job  AS C
                        JOIN graph_params AS GP
                        JOIN struct       AS S USING(struct_id)
                      WHERE (C.job_id,GP.graph_params_id)
                              NOT IN (SELECT G.job_id,G.graph_params_id FROM graph AS G)
                            AND C.bonddata IS NOT NULL
                      GROUP BY C.job_id,GP.graph_params_id"""

    ,template   = Temp(tempfuncs = [TempFunc(mk_graph
                                               ,args = noIndex(['atoms']+mk_graph_args[1:]))
                                   ,TempFunc(json_to_traj
                                               ,'atoms'
                                            ,args = [Arg('rawstruct')])
                                   ,TempFunc(graph_to_json
                                             ,'raw'
                                             ,args = [Arg('mk_graph')])])
    ,metatemp = SimpleInsert('graph',['job_id','graph_params_id','raw']))

h_bond = Rule('h_bonds'
    ,metatemp=PureSQL(["""
        UPDATE bond as B SET B.h_bond =
        IF(bondorder between 0.05 AND 0.4, (SELECT
                             (IF(A1.element_id < A2.element_id,
                                A1.element_id = 1 AND A2.element_id = 8,
                                A1.element_id = 8 AND A2.element_id = 1))

                        FROM  atom AS A1  JOIN atom AS A2
                            WHERE A1.struct_id = B.struct_id
                            AND A1.atom_id = B.atom1
                            AND A2.struct_id = B.struct_id
                            AND A2.atom_id = B.atom2),0)
            WHERE B.job_id>0 """]))

is_tom_job = Rule('is_tom'
    ,metatemp = PureSQL(["""
            UPDATE relax_job AS R SET R.is_tom =
                (SELECT J.stordir like '%%tom%%'
                        FROM alljob AS J
                        WHERE J.job_id = R.job_id)
            WHERE R.job_id > 0 """]))

rules = [get_chargemol_struct
             ,populate_bond
             ,set_default_graph
             ,populate_bonddata
             ,populate_graph
             ,h_bond
             ,is_tom_job]



from os import environ
from dbgen.support.datatypes.table     import Table,Col,FK
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.temp      import Temp,TempFunc,SimpleTemp
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.metatemp  import PureSQL,MetaTemp,Block
from dbgen.support.datatypes.rule      import Rule
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal

from dbgen.support.utils               import mkInsCmd
from dbgen.core.lists                  import flatten

from dbgen.scripts.Pure.Graph.json_to_graph  import json_to_graph
from dbgen.scripts.Pure.Graph.identify_ads   import identify_ads
##########################################################################################
##########################################################################################
##########################################################################################
# Constants
###########
pypth = environ['DBGEN_ROOT']+'/scripts/'
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']

ads = {'H':{1:1},'N':{7:1},'O':{8:1},'C':{6:1},'NH':{1:1,7:1},'OH':{8:1,1:1}
      ,'CH':{1:1,6:1},'NH2':{1:2,7:1},'H2O':{1:2,8:1},'CH2':{1:2,6:1}
      ,'NH3':{1:3,7:1},'CH3':{1:3,6:1},'N2':{7:2},'NNH':{1:1,7:2},'NNH2':{1:2,7:2}
      ,'NNH3':{1:3,7:2},'H2':{1:2},'C2':{6:2},'N2':{7:2},'O2':{8:2}
      ,'OOH':{1:1,8:2},'CHO':{1:1,6:1,8:1},'COOH':{1:1,6:1,8:2}
      ,'CO':{6:1,8:1},'OCCO':{6:2,8:2},'OCCHO':{1:1,6:2,8:2}
      }
##########################################################################################
##########################################################################################
##########################################################################################

functions = {
     'json_to_graph'    : pypth+'Pure/Graph/json_to_graph.py'
    ,'identify_ads'        : pypth+'Pure/Graph/identify_ads.py'

} # type: dict

##########################################################################################
##########################################################################################
##########################################################################################

adsorbate = Table('adsorbate'
    ,desc = 'Species that can adsorb onto a surface'
    ,cols = [Col('adsorbate_id',            pk = True, auto = True)
            ,Col('name',        Varchar(),  nn = True, uniq = True)
            ,Col('composition', Varchar())])

adsorbate_composition = Table('adsorbate_composition'
    ,'Components of an adsorbate'
    ,cols = [Col('adsorbate_id',  pk = True)
            ,Col('element_id',    pk = True)
            ,Col('number',        nn = True)])

struct_adsorbate = Table('struct_adsorbate'
    ,'An adsorbate considered on a particular surface'
    ,cols = [Col('struct_id',                   pk=True)
            ,Col('adsorbate_id',                pk=True)
            ,Col('struct_adsorbate_id',         pk=True)
            ,Col('site',                Varchar())]
    ,fks = [FK('struct_id',   'struct')
           ,FK('adsorbate_id','adsorbate')])

tables = [adsorbate
         ,adsorbate_composition
         ,struct_adsorbate]

##########################################################################################
##########################################################################################
##########################################################################################

ads_eng_view = View("ads_eng","""
SELECT R.ref_scheme_id                   -- SPECIFY *1* OF THESE
      ,A.adsorbate_id                    -- JOIN ON THIS
      ,J.calc_id                         -- JOIN ON THIS
      ,R.element_id                      -- SUM (number*avg_energy) WHEN GROUPING BY THIS
      ,R.component_id                    -- AVG energy IN SUBSELECT WHILE GROUPING BY THIS + Element&Component
      ,A.name                            -- human readable info
      ,E.symbol AS element               -- human readable info
      ,CE.symbol AS component            -- human readable info
      ,RS.struct_id                      -- join with struct for info to be used in constraints (e.g. kpt related things depending on bulk/molecule/surface)
      ,F.job_id                          -- join with JOB to be used in constraints (e.g. timestamp constraint to get repeatable results)
      ,F.energy/C.count AS energy_norm   -- contains fundamental information we want (averaged for each component element and summed to get adsorbate reference energy)
      ,J.calc_other_id                   -- join with calc other to filter down possible contributions to adsorbate energy
      ,AC.number                         -- how many of ELEMENT are in the adsorbate
      ,R.coef                            -- how much of each COMPONENT refjob goes into the calculation of the energy per ELEMENT atom
    FROM adsorbate_composition AS AC
    JOIN ref_scheme_reqs AS R ON AC.element_id = R.element_id
    JOIN adsorbate AS A USING (adsorbate_id)
    JOIN ref_scheme_refstruct AS RS ON RS.element_id=R.component_id AND RS.ref_scheme_id = R.ref_scheme_id
    JOIN finaltraj AS F ON F.struct_id = RS.struct_id
    JOIN relax_job AS J ON J.job_id = F.job_id
    JOIN element AS E ON E.element_id=R.component_id
    JOIN element AS CE ON CE.element_id=R.component_id
    JOIN struct_composition AS C ON C.element_id=RS.element_id AND C.struct_id = F.struct_id""")
views = [ads_eng_view]
##########################################################################################
##########################################################################################
##########################################################################################


adscompstr  = "((SELECT adsorbate_id FROM adsorbate WHERE name='%s'),%d,%d)"
adscompstrs = flatten([[adscompstr%(k,k2,v2) for k2,v2 in v.items()] for k,v in ads.items()])
populate_adsorbate = Rule('populate_adsorbate'
    ,metatemp = PureSQL(["""INSERT INTO adsorbate (name) VALUES %s
                            ON DUPLICATE KEY UPDATE name=name"""%','.join(["('%s')"%x for x in ads.keys()])
                       ,"""INSERT INTO adsorbate_composition (adsorbate_id,element_id,number) VALUES %s
                           ON DUPLICATE KEY UPDATE adsorbate_id=adsorbate_id"""%','.join(adscompstrs)]))


#####
struct_ads_cols = ['struct_id','adsorbate_id','struct_adsorbate_id','site']
pop_struct_ads = Rule('pop_struct_ads'
    ,query = """SELECT S.struct_id,S.geo_graph
                FROM struct AS S JOIN surface USING (struct_id)
                WHERE EXISTS (SELECT A.struct_adsorbate_id IS NULL
                                AND A.element_id IN (1,6,7,8)
                              FROM atom AS A
                              WHERE A.struct_id = S.struct_id)"""

    ,template = Temp(tempfuncs = [TempFunc(json_to_graph
                                          ,args=[Arg('geo_graph')])
                                 ,TempFunc(identify_ads
                                           ,name = 'ads'
                                           ,args = [Arg('json_to_graph')])])
    ,metatemp = MetaTemp([
                     Block(name = 'adsorbate_id'
                          ,text = 'SELECT A.adsorbate_id FROM adsorbate AS A WHERE A.name=%s'
                          ,args = [Arg('ads',4)])
                    ,Block(name='adsorbate_id_many'
                          ,text ='SELECT A.adsorbate_id FROM adsorbate AS A WHERE A.name=%s'
                          ,args = [Arg('ads',2)])
                    ,Block(name = 'insert_struct_ads'
                          ,text = mkInsCmd('struct_adsorbate',struct_ads_cols,sqlite=False)
                          ,args = [Arg('struct_id')
                                  ,Arg('adsorbate_id')
                                  ,Arg('ads',3)
                                  ,Arg('ads',5)])
                    ,Block(name = 'update_atom_fk'
                          ,text = """UPDATE atom SET struct_id=%s,adsorbate_id=%s,struct_adsorbate_id=%s
                                     WHERE struct_id=%s AND atom_id=%s"""
                          ,args = [Arg('struct_id') # needed because it's a composite foreign key?
                                  ,Arg('adsorbate_id_many')
                                  ,Arg('ads',1)
                                  ,Arg('struct_id')
                                  ,Arg('ads',0)])]))

pop_ads_compstr = Rule('pop_ads_compstr'
    ,metatemp=PureSQL(["""
            UPDATE adsorbate AS A SET A.composition=
              (select  GROUP_CONCAT(CONCAT('[',C.element_id,',',C.number,']') ORDER BY C.element_id ASC)
                    FROM adsorbate_composition as C
                    WHERE C.adsorbate_id = A.adsorbate_id)"""]))

rules = [populate_adsorbate
             ,pop_struct_ads
             ,pop_ads_compstr]



from os import environ
from dbgen.support.datatypes.rule import Rule
from dbgen.support.datatypes.temp      import SimpleTemp,Const
from dbgen.support.datatypes.metatemp  import SimpleInsert,MetaTemp,Block,noIndex
from dbgen.support.datatypes.misc      import Arg
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal

from dbgen.support.utils               import mkInsCmd
from dbgen.support.datatypes.table      import Table,Col,FK
from dbgen.scripts.IO.parse_keld import parse_keld
pypth = environ['DBGEN_ROOT']+'/scripts/'

##########################################################################################
keld_table = Table('keld_solids'
    ,desc='Data from Keld L. with experimental properties for bulk materials'
    ,cols = [Col('id',                            pk=True,auto=True)
            ,Col('name',             Varchar(),       nn=True)
            ,Col('job_name',         Varchar(),       nn=True)
            ,Col('lattice_parameter',Decimal())
            ,Col('cohesive_energy',  Decimal())
            ,Col('bulkmod',          Decimal())
            ,Col('structure',        Varchar())
            ,Col('magmom',           Decimal())])

tables = [keld_table]

##########################################################################################
##########################################################################################
##########################################################################################

keld_cols2 = ['name_short','pure_struct_id','spacegroups'
             ,'name_mid','element_id','count'
             ,'name_long','property','value']

s_d_cols = ['struct_dataset_id','species_id','property','value']

pop_keld = Rule('keld'
    ,template = SimpleTemp(parse_keld,[],keld_cols2
                          ,{'k':Const('keld_solids')})
    ,metatemp = MetaTemp([
         Block(name = 'insert_species'
              ,text = mkInsCmd('species',['name','nickname'])
              ,args = noIndex(['name_short','name_short']))

        ,Block(name = 'insert_pure_struct'
              ,text = mkInsCmd('pure_struct',['name','spacegroup'])
              ,args = noIndex(['pure_struct_id','spacegroups']))

        ,Block(name = 'query_species_short'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_short')])

        ,Block(name = 'query_pure_struct'
              ,text = 'SELECT pure_struct_id FROM pure_struct WHERE name=%s'
              ,args = [Arg('pure_struct_id')])

        ,Block(name = 'insert_bulk_species'
              ,text = mkInsCmd('bulkspecies',['species_id','pure_struct_id'],sqlite=False)
              ,args = [Arg('query_species_short')
                      ,Arg('query_pure_struct')])

        ,Block(name = 'query_species_mid'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_mid')])

        ,Block(name = 'insert_species_comp'
              ,text = mkInsCmd('species_composition',['species_id','element_id','count'],sqlite=False)
              ,args = [Arg('query_species_mid')
                      ,Arg('element_id')
                      ,Arg('count')])

        ,Block(name = 'query_species_long'
              ,text = 'SELECT species_id FROM species WHERE name=%s'
              ,args = [Arg('name_long')])

        ,Block(name = 'insert_dataset'
              ,text = """INSERT INTO struct_dataset (name) VALUES (%s)
                      ON DUPLICATE KEY UPDATE struct_dataset_id=struct_dataset_id"""
              ,args = [Arg('k')])

        ,Block(name = 'query_dataset'
              ,text = """SELECT  struct_dataset_id FROM struct_dataset
                        WHERE name = %s"""
              ,args = [Arg('k')])

        ,Block(name = 'insert_dataset_elem'
              ,text = mkInsCmd('struct_dataset_element',s_d_cols,sqlite=False)
              ,args = [Arg('query_dataset')
                      ,Arg('query_species_long')
                      ,Arg('property'),Arg('value')])]))

rules = [pop_keld]



from typing import Callable,TypeVar,Optional
from  ase import Atoms # type: ignore
from glob import glob
from ase.io import read # type: ignore
from os.path import join,exists,getsize
from numpy.linalg import norm # type: ignore
from numpy import vectorize   # type: ignore

"""
Miscellaneous Core Functions / Objects

higher order
    maybe
    cast_maybe
    identity

numeric
    roundfloat

IO
    anytraj

Strings
    levenshteinDistance
"""

################################################################################
# Higher order functions
#------------------
X = TypeVar('X')
Y = TypeVar('Y')
def maybe(f : Callable[[X],Y]
         ) -> Callable[[Optional[X]]
                          ,Optional[Y]]:
    """
    Lift a unary function (that might normally fail on a None input)
        such that it only applies the function if the argument is not None
        (returns None if it is)
    """
    def g(x : Optional[X])->Optional[Y]:
        if x is None:
            return None
        else:
            return f(x)
    return g

# Dict[str,Callable]
cast_maybe = {'int':  maybe(int)  # type: ignore
              ,'bool': maybe(lambda x: 1 if '.TRUE.' in x else 0) # type: ignore
              ,'str' : maybe(lambda x: x)  # type: ignore
              ,'float': maybe(float) # type: ignore
              }

T = TypeVar('T')
def identity(x:T)->T:
    return x # type: ignore

A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')

def compose(x : A
           ,f : Callable[[A],B]  = identity
           ,g : Callable[[B],C]  = identity
           ) -> C:
    return g(f(x))
################################################################################
################################################################################
def anytraj(root : str
            ) -> Atoms:
    """
    ASE IO read function - takes any traj it can find
    """
    qn = join(root,'qn.traj')
    if exists(qn) and getsize(qn) > 100:
        return read(qn)
    trajs = glob(join(root,'*.traj'))
    if len(trajs)==0:
        raise ValueError('Get Traj could not find any traj in '+root)
    else:
        return read(trajs[0])

def wrap(atoms:Atoms)->Atoms:
    mags = [norm(x) for x in atoms.get_cell()]
    def wrap_coord(coord:float,cell_len:float)->float:
        if coord < 0: coord+=cell_len
        elif coord > cell_len: coord-=cell_len
        return coord
    for i in range(len(atoms)):
        for j in range(3):

            atoms.positions[i][j] = wrap_coord(atoms.positions[i][j],mags[j]) # atoms.positions[i][j] =
            if atoms.positions[i][j] < 0:
                import pdb;pdb.set_trace()
    return atoms
################################################################################
################################################################################

def levenshteinDistance(s1 : str, s2 : str) -> int:
    if len(s1) > len(s2):
        s1, s2 = s2, s1

    distances = list(range(len(s1) + 1))
    for i2, c2 in enumerate(s2):
        distances_ = [i2+1]
        for i1, c1 in enumerate(s1):
            if c1 == c2:
                distances_.append(distances[i1])
            else:
                distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
        distances = distances_
    return distances[-1]



from dbgen.support.datatypes.object     import Object,Attr,Universe
from dbgen.support.datatypes.sqltypes   import SQLType,Int,Varchar,Text,Decimal
from dbgen.support.datatypes.constraint import *
"""
Object-based model, gets processed into Tables/Rules
"""
################################################################################
alljob = Object(name='alljob'
    ,desc = 'DFT jobs'
    ,attr = [Attr('logfile',          Varchar()       ,nnull=True)
            ,Attr('stordir',          Varchar()       ,nnull=True)
            ,Attr('code',             Varchar()       ,nnull=True)
            ,Attr('user',             Varchar())
            ,Attr('timestamp')
            ,Attr('working_directory',Varchar())
            ,Attr('log',              Text('long'))
            ,Attr('pwinp',            Text())
            ,Attr('potcar',           Text('long'))
            ,Attr('poscar',           Text())
            ,Attr('kptcar',           Text())
            ,Attr('paramdict',        Text('long'))
            ,Attr('anytraj',          Text('long'))
            ,Attr('job_type',         Varchar()       ,ind=True)
            ,Attr('job_name',         Varchar())
            ,Attr('ads_catalog',      Varchar())
            ,Attr('structure_catalog',Varchar())
            ,Attr('deleted', nnull=True,ind=True,default=0)]
    ,uniqattr = ['logfile'])

calc = Object(name = 'calc'
    ,desc = 'DFT calc parameters'
    ,attr = [Attr('dftcode',   Varchar(), nnull = True)
            ,Attr('xc',        Varchar(), nnull = True)
            ,Attr('pw',                   nnull = True)
            ,Attr('psp',       Varchar(), nnull = True)]
    ,uniqattr = ['dftcode','xc','pw','psp'])

calc_other= Object('calc_other'
    ,desc = 'Less important DFT Calculator Parameters'
    ,attr = [Attr("kx")
            ,Attr("ky")
            ,Attr("kz")
            ,Attr("fmax",            Decimal())
            ,Attr("econv",           Decimal(10,7))
            ,Attr("dw")
            ,Attr("sigma",           Decimal(10,7))
            ,Attr("nbands")
            ,Attr("mixing",          Decimal())
            ,Attr("nmix")
            ,Attr("xtol",            Decimal(10,7))
            ,Attr("strain",          Decimal())
            ,Attr("gga",             Varchar())
            ,Attr("luse_vdw")
            ,Attr("zab_vdw",         Decimal())
            ,Attr("nelmdl")
            ,Attr("gamma")
            ,Attr("dipol",           Varchar())
            ,Attr("algo",            Varchar())
            ,Attr("ibrion")
            ,Attr("prec",            Varchar())
            ,Attr("ionic_steps")
            ,Attr("lreal",           Varchar())
            ,Attr("lvhar")
            ,Attr("diag",            Varchar())
            ,Attr("spinpol")
            ,Attr("dipole")
            ,Attr("maxstep")
            ,Attr("delta",           Decimal())
            ,Attr("mixingtype",      Varchar())
            ,Attr("bonded_inds",     Varchar())
            ,Attr("energy_cut_off",  Decimal())
            ,Attr("step_size",       Decimal())
            ,Attr("spring",          Decimal())
            ,Attr('cell_dofree',     Varchar())
            ,Attr('cell_factor',     Decimal())
            ,Attr("kpts",            Varchar())])

relax_job = Object(name = 'relax_job'
    ,desc       = 'Jobs that compute local minima for electronic energy'
    ,parents    = ['alljob']
    ,components = ['calc','calc_other']
    ,attr       = [Attr('reference')
                  ,Attr('is_tom')])

roots = Object('roots'
    ,desc = 'Directories which are recursively searched for computation logfiles'
    ,attr = [Attr('root', Varchar(),nnull=True)
            ,Attr('code', Varchar(),nnull=True)
            ,Attr('label',Varchar(),nnull=True)
            ,Attr('active',         nnull=True,default=1)]
    ,uniqattr = ['root','code','label'])


cell = Object('cell'
    ,desc = 'Periodic cells defined by three vectors'
    ,attr = [Attr('a0',          Decimal(),nnull = True)
            ,Attr('a1',          Decimal(),nnull = True)
            ,Attr('a2',          Decimal(),nnull = True)
            ,Attr('b0',          Decimal(),nnull = True)
            ,Attr('b1',          Decimal(),nnull = True)
            ,Attr('b2',          Decimal(),nnull = True)
            ,Attr('c0',          Decimal(),nnull = True)
            ,Attr('c1',          Decimal(),nnull = True)
            ,Attr('c2',          Decimal(),nnull = True)
            ,Attr('surface_area',Decimal())
            ,Attr('volume',      Decimal())
            ,Attr('a',           Decimal())
            ,Attr('b',           Decimal())
            ,Attr('c',           Decimal())]
    ,uniqattr= ['a0','a1','a2','b0','b1','b2','c0','c1','c2'])


pure_struct = Object('pure_struct'
    ,desc = 'Structure abstraction developed by Ankit Jain'
    ,attr = [Attr('name',            Varchar(),  nnull = True)
            ,Attr('spacegroup')
            ,Attr('free')
            ,Attr('nickname',        Varchar())
            ]
    ,uniqattr = ['name'])

species = Object('species'
    ,desc="""A combination of composition and structural information (no floats)"""
    ,attr=[Attr('name',          Varchar())
          ,Attr('nickname',      Varchar())
          ,Attr('composition',   Varchar())]
    ,uniqattr = ['name'])


struct = Object('struct'
    ,desc       = 'List of unique combinations of [atom] and cell'
    ,components = ['cell','pure_struct','pure_struct','species']
    ,attr = [Attr('raw',             Text(),    nnull  = True)
            ,Attr('rawhash',         Varchar(), nnull  = True)
            ,Attr('system_type',     Varchar(), ind    = True) # used to partition subclasses
            ,Attr('n_atoms')
            ,Attr('n_elems')
            ,Attr('composition',     Varchar())
            ,Attr('composition_norm',Varchar())
            ,Attr('metal_comp',      Varchar())
            ,Attr('str_symbols',     Varchar())
            ,Attr('str_constraints', Varchar())
            ,Attr('symmetry',        Varchar())
            ,Attr('pure_struct_id_catalog')
            ,Attr('spacegroup')
            ,Attr('geo_graph',       Text('long'))
            ,Attr('elemental')]
        ,uniqattr= ['rawhash'])


traj = Object('traj'
    ,desc       = 'A step in a relaxation'
    ,parents    = ['alljob']
    ,components = ['struct']
    ,attr = [Attr('final',                   nnull  = True)
            ,Attr('energy',        Decimal(),ind    = True)
            ,Attr('fmax',          Decimal())
            ,Attr('kptden_x',      Decimal())
            ,Attr('kptden_y',      Decimal())
            ,Attr('kptden_z',      Decimal())]
    ,many_to_one = True)

element = Object('element'
    ,desc = 'ID = atomic number' # table for storing data in element.json
    ,attr = [Attr('symbol',                 Varchar(),  nnull = True)
            ,Attr('atomic_weight',          Decimal(),nnull = True)
            ,Attr('name',                   Varchar(),  nnull = True)
            ,Attr('atomic_radius')
            ,Attr('phase',                  Varchar())
            ,Attr('group_id')
            ,Attr('period')
            ,Attr('pointgroup',             Varchar())
            ,Attr('spacegroup')
            ,Attr('evaporation_heat',        Decimal())
            ,Attr('melting_point',           Decimal())
            ,Attr('metallic_radius',         Decimal())
            ,Attr('vdw_radius',              Decimal())
            ,Attr('density',                 Decimal())
            ,Attr('en_allen',                Decimal())
            ,Attr('is_radioactive')
            ,Attr('lattice_struct',          Varchar())
            ,Attr('fusion_heat',             Decimal())
            ,Attr('econf',                   Varchar())
            ,Attr('covalent_radius_bragg',   Decimal())
            ,Attr('geochemical_class',       Varchar())
            ,Attr('abundance_crust',         Decimal())
            ,Attr('heat_of_formation',       Decimal())
            ,Attr('electron_affinity',       Decimal())
            ,Attr('atomic_volume',           Decimal())
            ,Attr('boiling_point',           Decimal())
            ,Attr('proton_affinity',         Decimal())
            ,Attr('covalent_radius_slater',  Decimal())
            ,Attr('lattice_constant',        Decimal())
            ,Attr('dipole_polarizability',   Decimal())
            ,Attr('en_ghosh',                Decimal())
            ,Attr('thermal_conductivity',    Decimal())
            ,Attr('en_pauling',              Decimal())
            ,Attr('gas_basicity',            Decimal())
            ,Attr('abundance_sea',           Decimal())])

adsorbate = Object('adsorbate'
    ,desc     = 'Species that can adsorb onto a surface'
    ,attr     = [Attr('name',        Varchar(),  nnull = True)
                ,Attr('composition', Varchar())]
    ,uniqattr = ['name'])

adsorbate_composition = Object('adsorbate_composition'
    ,'Components of an adsorbate'
    ,parents = ['adsorbate','element']
    ,attr    = [Attr('number', nnull = True)])

struct_adsorbate = Object('struct_adsorbate'
    ,desc        = 'An adsorbate considered on a particular surface'
    ,parents     = ['struct','adsorbate']
    ,many_to_one = True
    ,attr        = [Attr('site', Varchar())])

atom = Object('atom'
    ,desc        = 'List of atoms in unique structs'
    ,parents     = ['struct']
    ,many_to_one = True
    ,components  = ['element','struct_adsorbate']
    ,attr = [Attr('x',            Decimal(),    nnull = True)
            ,Attr('y',            Decimal(),    nnull = True)
            ,Attr('z',            Decimal(),    nnull = True)
            ,Attr('constrained')
            ,Attr('magmom',       Decimal())
            ,Attr('tag')
            ,Attr('v3'),Attr('v4'),Attr('v5'),Attr('v6')
            ,Attr('v7'),Attr('v8'),Attr('v9'),Attr('v10')
            ,Attr('layer')])


chargemol_job = Object('chargemol_job'
        ,desc = 'Jobs that compute bond order analysis with Chargemol'
        ,parents = ['alljob']
        ,components= ['calc','struct']
        ,attr = [Attr('bonddata',Text('long'))])

bond = Object('bond'
    ,desc        = "A bond between two atoms identified by chargemol"
    ,parents     = ['chargemol_job','atom1','atom2']
    ,many_to_one =  True
    ,attr = [Attr('bondorder',Decimal(), nnull = True)
            ,Attr('distance', Decimal(), nnull = True)
            ,Attr('pbc_x',               nnull = True)
            ,Attr('pbc_y',               nnull = True)
            ,Attr('pbc_z',               nnull = True)
            ,Attr('h_bond')])

uni = Universe([alljob
                ,relax_job
                ,roots
                ,calc
                ,calc_other
                ,traj
                ,struct
                ,cell
                ,element
                ,atom
                ,adsorbate
                ,adsorbate_composition
                ,struct_adsorbate
                ,pure_struct
                ,species
                ,chargemol_job

                ])



from os import environ
from dbgen.support.datatypes.table     import Table,Col,FK,View
from dbgen.support.datatypes.temp      import Temp,TempFunc,SimpleTemp,Const
from dbgen.support.datatypes.misc      import Arg,noIndex
from dbgen.support.datatypes.metatemp  import SimpleUpsert,SimpleUpdate,SimpleInsert,PureSQL,MetaTemp,Block
from dbgen.support.datatypes.rule      import Rule
from dbgen.support.utils               import addQs,mkInsCmd
from dbgen.core.lists                  import flatten
from dbgen.support.datatypes.sqltypes  import Varchar,Decimal,Text

from dbgen.scripts.IO.get_gpaw_logfiles         import get_gpaw_logfiles
from dbgen.scripts.IO.get_qe_logfiles           import get_qe_logfiles
from dbgen.scripts.IO.get_catalog_logfiles      import get_catalog_logfiles
from dbgen.scripts.IO.get_chargemol_logfiles    import get_chargemol_logfiles
from dbgen.scripts.IO.anytraj                   import anytraj
from dbgen.scripts.IO.parse_mendeleev           import parse_mendeleev
from dbgen.scripts.IO.get_job_metadata          import get_job_metadata
from dbgen.scripts.Pure.Load.parse_pw_gpaw      import parse_pw_gpaw
from dbgen.scripts.Pure.Load.parse_pw_qe        import parse_pw_qe
from dbgen.scripts.Pure.Load.parse_pw_vasp      import parse_pw_vasp
from dbgen.scripts.Pure.Load.parse_xc_gpaw      import parse_xc_gpaw
from dbgen.scripts.Pure.Load.parse_xc_qe        import parse_xc_qe
from dbgen.scripts.Pure.Load.parse_xc_vasp      import parse_xc_vasp
from dbgen.scripts.Pure.Load.parse_psp_gpaw     import parse_psp_gpaw
from dbgen.scripts.Pure.Load.parse_psp_qe       import parse_psp_qe
from dbgen.scripts.Pure.Load.parse_psp_vasp     import parse_psp_vasp
from dbgen.scripts.Pure.Load.normalize_xc       import normalize_xc
from dbgen.scripts.Pure.Load.normalize_psp      import normalize_psp
from dbgen.scripts.Pure.Atoms.get_traj_gpaw_lazy import get_traj_gpaw_lazy
from dbgen.scripts.Pure.Atoms.get_traj_qe_lazy   import get_traj_qe_lazy
from dbgen.scripts.Pure.Atoms.get_traj_vasp_lazy import get_traj_vasp_lazy
from dbgen.scripts.Pure.Misc.dftcode_dependent  import dftcode_dependent
from dbgen.scripts.Pure.Misc.triple             import triple
from dbgen.scripts.Pure.Misc.identity           import identity
from dbgen.scripts.Pure.Misc.unzip              import unzip
from dbgen.scripts.Pure.Misc.map_               import map_
from dbgen.scripts.Pure.Misc.concat_map         import concat_map
from dbgen.scripts.Pure.Atoms.json_to_traj      import json_to_traj
from dbgen.scripts.Pure.Atoms.traj_to_json      import traj_to_json
from dbgen.scripts.Pure.Atoms.get_cell          import get_cell
from dbgen.scripts.Pure.Atoms.get_atom          import get_atom
from dbgen.scripts.Pure.Atoms.get_system_type   import get_system_type
##########################################################################################
##########################################################################################
##########################################################################################
# CONSTANTS
###########

calc_funcs = [parse_pw_gpaw,parse_pw_qe,parse_pw_vasp # type: ignore
             ,parse_xc_gpaw,parse_xc_qe,parse_xc_vasp
             ,parse_psp_gpaw,parse_psp_qe,parse_psp_vasp
             ,normalize_xc,normalize_psp,identity]

traj_funcs = [get_cell,get_atom,traj_to_json,get_traj_gpaw_lazy,get_traj_qe_lazy
             ,get_traj_vasp_lazy,identity,unzip]

cell_args = [Arg('cell',i) for i in range(9)]
cell_cols = ['a0','a1','a2','b0','b1','b2','c0','c1','c2']

##########################################################################################
roots = Table('roots'
    ,desc='Directories which are recursively searched for computation logfiles'
    ,cols = [Col('storage_id',      pk=True,auto=True)
            ,Col('root',Varchar(),  nn=True,uniq=True)
            ,Col('code',Varchar(),  nn=True,uniq=True)
            ,Col('label',Varchar(), nn=True,uniq=True)
            ,Col('active',          nn=True)])

alljob = Table('alljob'
    ,desc='DFT jobs'
    ,cols = [Col('job_id',                          pk=True,auto=True)
            ,Col('logfile',          Varchar(),         nn=True,uniq=True)
            ,Col('stordir',          Varchar(),         nn=True)
            ,Col('code',             Varchar(),         nn=True)
            ,Col('user',             Varchar())
            ,Col('timestamp')
            ,Col('working_directory',Varchar())

            ,Col('log',              Text('long'))
            ,Col('pwinp',            Text())
            ,Col('potcar',           Text('long'))
            ,Col('poscar',           Text())
            ,Col('kptcar',           Text())
            ,Col('paramdict',        Text('long'))
            ,Col('anytraj',          Text('long'))

            ,Col('job_type',         Varchar(),         ind = True)
            ,Col('job_name',         Varchar())
            ,Col('ads_catalog',      Varchar())
            ,Col('structure_catalog',Varchar())
            ,Col('deleted',                         ind=True,nn=True,default=0)])


relax_job = Table('relax_job'
        ,desc = 'Jobs that compute local minima for electronic energy'
        ,cols = [Col('job_id',        pk  = True)
                ,Col('calc_id')
                ,Col('calc_other_id')
                ,Col('reference',     ind = True)
                ,Col('is_tom')]
        ,fks  = [FK('job_id', 'alljob')
                ,FK('calc_id','calc')
                ,FK('calc_other_id','calc_other')])

traj = Table('traj'
    ,desc = 'A step in a relaxation'
    ,cols = [Col('job_id',                pk  = True)
            ,Col('traj_id',               pk  = True)
            ,Col('final',                 nn  = True)
            ,Col('struct_id',             nn  = True)
            ,Col('energy',        Decimal(),ind = True)
            ,Col('fmax',          Decimal())
            ,Col('kptden_x',      Decimal())
            ,Col('kptden_y',      Decimal())
            ,Col('kptden_z',      Decimal())]
    ,fks = [FK('job_id',   'relax_job')
           ,FK('struct_id','struct')])

#########################################
calc = Table('calc'
    ,desc='DFT calc parameters'
    ,cols = [Col('calc_id',          pk = True,auto = True)
            ,Col('dftcode',   Varchar(), nn = True,uniq = True)
            ,Col('xc',        Varchar(), nn = True,uniq = True)
            ,Col('pw',                   nn = True,uniq = True)
            ,Col('psp',       Varchar(), nn = True,uniq = True)])

element = Table('element'
    ,desc = 'ID = atomic number' # table for storing data in element.json
    ,cols = [Col('element_id',                     pk = True)
            ,Col('symbol',                 Varchar(),  nn = True)
            ,Col('atomic_weight',          Decimal(),nn = True)
            ,Col('name',                   Varchar(),  nn = True)
            ,Col('atomic_radius')
            ,Col('phase',                  Varchar())
            ,Col('group_id')
            ,Col('period')
            ,Col('pointgroup',             Varchar())
            ,Col('spacegroup')
            ,Col('evaporation_heat',        Decimal())
            ,Col('melting_point',           Decimal())
            ,Col('metallic_radius',         Decimal())
            ,Col('vdw_radius',              Decimal())
            ,Col('density',                 Decimal())
            ,Col('en_allen',                Decimal())
            ,Col('is_radioactive')
            ,Col('lattice_struct',          Varchar())
            ,Col('fusion_heat',             Decimal())
            ,Col('econf',                   Varchar())
            ,Col('covalent_radius_bragg',   Decimal())
            ,Col('geochemical_class',       Varchar())
            ,Col('abundance_crust',         Decimal())
            ,Col('heat_of_formation',       Decimal())
            ,Col('electron_affinity',       Decimal())
            ,Col('atomic_volume',           Decimal())
            ,Col('boiling_point',           Decimal())
            ,Col('proton_affinity',         Decimal())
            ,Col('covalent_radius_slater',  Decimal())
            ,Col('lattice_constant',        Decimal())
            ,Col('dipole_polarizability',   Decimal())
            ,Col('en_ghosh',                Decimal())
            ,Col('thermal_conductivity',    Decimal())
            ,Col('en_pauling',              Decimal())
            ,Col('gas_basicity',            Decimal())
            ,Col('abundance_sea',           Decimal())])
#########
cell = Table('cell'
    ,desc = 'Periodic cells defined by three vectors'
    ,cols = [Col('cell_id',             pk = True,auto = True)
            ,Col('a0',          Decimal(),nn = True,uniq = True)
            ,Col('a1',          Decimal(),nn = True,uniq = True)
            ,Col('a2',          Decimal(),nn = True,uniq = True)
            ,Col('b0',          Decimal(),nn = True,uniq = True)
            ,Col('b1',          Decimal(),nn = True,uniq = True)
            ,Col('b2',          Decimal(),nn = True,uniq = True)
            ,Col('c0',          Decimal(),nn = True,uniq = True)
            ,Col('c1',          Decimal(),nn = True,uniq = True)
            ,Col('c2',          Decimal(),nn = True,uniq = True)
            ,Col('surface_area',Decimal())
            ,Col('volume',      Decimal())
            ,Col('a',           Decimal())
            ,Col('b',           Decimal())
            ,Col('c',           Decimal())])

struct = Table('struct'
    ,desc = 'List of unique combinations of [atom] and cell'
    ,cols = [Col('struct_id',                pk  = True,auto = True)
            ,Col('cell_id',                  nn  = True)
            ,Col('raw',             Text(),  nn  = True)
            ,Col('rawhash',         Varchar(),   nn  = True,uniq = True)
            ,Col('system_type',     Varchar(),   ind = True) # used to partition subclasses
            ,Col('n_atoms')
            ,Col('n_elems')
            ,Col('composition',     Varchar())
            ,Col('composition_norm',Varchar())
            ,Col('metal_comp',      Varchar())
            ,Col('str_symbols',     Varchar())
            ,Col('str_constraints', Varchar())
            ,Col('symmetry',        Varchar())
            ,Col('pure_struct_id')
            ,Col('pure_struct_id_catalog')
            ,Col('species_id')
            ,Col('spacegroup')
            ,Col('geo_graph',       Text('long'))
            ,Col('elemental')]
    ,fks  = [FK('cell_id',       'cell')
            ,FK('pure_struct_id','pure_struct')
            ,FK('species_id','species')])

atom = Table('atom'
    ,desc = 'List of atoms in unique structs'
    ,cols = [Col('struct_id',                pk = True)
            ,Col('atom_id',                  pk = True)
            ,Col('x',            Decimal(),  nn = True)
            ,Col('y',            Decimal(),  nn = True)
            ,Col('z',            Decimal(),  nn = True)
            ,Col('constrained')
            ,Col('magmom',       Decimal())
            ,Col('tag')
            ,Col('v3')
            ,Col('v4')
            ,Col('v5')
            ,Col('v6')
            ,Col('v7')
            ,Col('v8')
            ,Col('v9')
            ,Col('v10')
            ,Col('adsorbate_id')
            ,Col('struct_adsorbate_id')
            ,Col('adsorbate_id_catalog')
            ,Col('struct_adsorbate_id_catalog')
            ,Col('layer')]
    ,fks = [FK('struct_id', 'struct')
           ,FK('element_id','element')
           ,FK(['struct_id'
               ,'adsorbate_id'
               ,'struct_adsorbate_id']
               ,'struct_adsorbate')])


trajatom = Table('trajatom'
        ,desc = "An atom considered as part of an optimization trajectory"
        ,cols = [Col('job_id',        pk=True)
                ,Col('traj_id',        pk=True)
                ,Col('struct_id',    pk=True)
                ,Col('atom_id',        pk=True)
                ,Col('fx',Decimal())
                ,Col('fy',Decimal())
                ,Col('fz',Decimal())]
        ,fks  = [FK(['job_id','traj_id'],'traj')
                ,FK(['struct_id','atom_id'],'atom')])
tables = [roots,alljob,relax_job,traj,calc,element,cell,struct,atom,trajatom]
##########################################################################################
##########################################################################################
##########################################################################################
job = View('job',"""SELECT A.job_id,A.user,A.timestamp,A.logfile,A.stordir,A.code
                          ,A.job_type,A.working_directory,A.ads_catalog,A.structure_catalog
                          ,A.deleted,A.job_name
                    FROM alljob AS A
                    WHERE NOT A.deleted""")
finaltraj = View('finaltraj','SELECT * FROM traj WHERE traj.final')


views = [job,finaltraj]
##########################################################################################
##########################################################################################
##########################################################################################

populate_gpaw_job = Rule(name = 'stray_gpaw_logfiles'
    ,query    = """SELECT R.root
                    FROM roots AS R
                    WHERE R.code = 'gpaw'
                      AND R.label = 'stray'
                      AND R.active
                     GROUP BY R.root"""
    ,template = SimpleTemp(get_gpaw_logfiles
                          ,['root']
                          ,['stordir','logfile']
                          ,{'code':Const('gpaw')})
    ,metatemp = SimpleInsert('alljob',['stordir','logfile','code']))

populate_qe_job = Rule(name = 'stray_qe_logfiles'
    ,query    = """SELECT R.root
                    FROM roots AS R
                    WHERE R.code = 'quantumespresso'
                      AND R.label = 'stray'
                      AND R.active"""
    ,template = SimpleTemp(get_qe_logfiles
                          ,['root']
                          ,['stordir','logfile']
                          ,{'code':Const('quantumespresso')})
    ,metatemp = SimpleInsert('alljob',['stordir','logfile','code']))

pop_job_cols = ['stordir','logfile','code','log','pwinp','potcar','poscar','kptcar'
               ,'paramdict']
# GROUP concat string is TOO BIG (and it's much faster without checking)
populate_catalog_job = Rule(name = 'catalog'
    ,query    = """SELECT  R.root
                    FROM roots AS R
                    WHERE R.label = 'catalog'
                      AND R.active
                    """
    ,template = SimpleTemp(get_catalog_logfiles
                          ,['root']
                          ,pop_job_cols)
    ,metatemp = SimpleInsert('alljob',pop_job_cols))

populate_chgmol_job = Rule(name = 'stray_chargemol_logfiles'
    ,query    = """SELECT  R.root
                    FROM roots AS R
                    WHERE R.code = 'chargemol'
                        AND R.label = 'stray'
                        AND R.active
                    """
    ,template = SimpleTemp(get_chargemol_logfiles
                          ,['root']
                          ,['stordir','logfile']
                          ,{'code':Const('chargemol')})
    ,metatemp = SimpleInsert('alljob',['stordir','logfile','code']))

anyTraj = Rule(name='anytraj'
    ,query    = """SELECT J.stordir,J.job_id FROM alljob AS J
                    WHERE J.anytraj is NULL AND NOT J.deleted"""
    ,template = SimpleTemp(anytraj,['stordir'],['anytraj'])
    ,metatemp = SimpleUpdate(alljob,['anytraj']))
################################################################################
mdata_cols = ['user','timestamp','working_directory'
             ,'job_type','job_name']

getJobMetadata = Rule(name = 'metadata'
    ,query    = """SELECT J.job_id,J.stordir,J.code,J.paramdict
                    FROM alljob AS J
                    WHERE J.user is null AND NOT J.deleted"""
    ,template = SimpleTemp(get_job_metadata,['stordir','code','paramdict'],mdata_cols)
    ,metatemp = SimpleUpdate(tab  = alljob
                            ,cols = mdata_cols))

################################################################################

populate_relax_calc = Rule(name='calc'
        ,query    = """SELECT relax_job.job_id,J.log,J.code AS dftcode
                        FROM relax_job
                        JOIN alljob as J USING (job_id)
                        WHERE relax_job.calc_id is null
                             AND NOT J.deleted"""
        ,template = Temp(
            constants = {f.__name__:Const(f) for f in calc_funcs}
            ,tempfuncs = [TempFunc(dftcode_dependent
                                      ,name = x
                                      ,args = noIndex(['dftcode'
                                                      ,'log'
                                                      ,'parse_%s_gpaw'%x
                                                      ,'parse_%s_qe'%x
                                                      ,'parse_%s_vasp'%x
                                                      ,'identity'
                                                      ,'identity' if x =='pw' else 'normalize_%s'%x]))
                                                      for x in ['pw','xc','psp']])

        ,metatemp = SimpleUpsert(ins_tab   = calc
                                ,ins_cols  = ["dftcode","xc","pw","psp"]
                                ,fk_tab    = relax_job))
populate_relax_job = Rule(name = 'relax_job'
        ,metatemp = PureSQL([
        """ INSERT INTO relax_job (job_id)
                SELECT J.job_id FROM alljob AS J
                    WHERE J.job_type IN ('relax','latticeopt','vcrelax')
                    AND J.job_id NOT IN (SELECT R.job_id from relax_job AS R)
                    AND NOT J.deleted"""]))



# if any ionic steps are in the traj table, then we assume all 4 tables have been taken care of
populate_traj = Rule(name='traj'
    ,query = """SELECT R.job_id,J.code AS dftcode,J.log,J.anytraj,J.poscar,J.stordir
                 FROM relax_job AS R JOIN alljob AS J USING (job_id)
                 WHERE R.job_id NOT IN (SELECT traj.job_id FROM traj)
                 AND J.code != 'vasp' # TEMPORARY THING
                 AND NOT J.deleted"""

    ,template = Temp(constants = {f.__name__:Const(f) for f in traj_funcs}
        ,tempfuncs=[
              TempFunc(triple
                      ,'anytraj_log_poscar'
                      ,args = noIndex(['anytraj','log','poscar']))


              ,TempFunc(dftcode_dependent
                       ,name = 'get_traj'
                       ,args = noIndex(['dftcode'
                                  ,'anytraj_log_poscar'
                                  ,'get_traj_gpaw_lazy'
                                  ,'get_traj_qe_lazy'
                                  ,'get_traj_vasp_lazy'
                                  ,'identity'
                                  ,'unzip'])) # returns ([ASEATOMS]m,[STEPNUM]m,[ENERGY]m,[TRAJATOMLIST]m)
               ,TempFunc(map_
                        ,name = 'cell'
                        ,args = noIndex(['get_cell','structs_few'])) # returns ([ax]m,[ay]m,...)
               ,TempFunc(map_
                        ,name = 'structs_few'
                        ,args = [Arg('traj_to_json')
                                ,Arg('get_traj', 0)]) # returns [raw_json]m
               ,TempFunc(concat_map
                        ,name = 'atoms'
                        ,args = noIndex(['get_atom','structs_few'])) # returns ([ASEATOMS]n,[Ind]n,[ELEMENT]n,[X]n,[Y]n,...)
               ,TempFunc(identity
                        ,name = 'structs_many'
                        ,args = [Arg('atoms', 0)])# returns [raw_json]n
                ,TempFunc(flatten
                         ,name = 'flattened_trajatoms'
                         ,args = [Arg('get_traj',3)])  # [STEPNUM,Ind,FX,FY,FZ]n
                ,TempFunc(unzip
                         ,'trajatoms'
                         ,args=[Arg('flattened_trajatoms')])]) # ([STEPNUM]n,[Ind]n,[FX]n,[FY]n,[FZ]n)

    ,metatemp = MetaTemp([
        Block(name = 'insert_cell'
             ,text = mkInsCmd('cell',cell_cols,sqlite=False)
             ,args = cell_args)

       ,Block(name = 'query_cell'
             ,text = ('SELECT C.cell_id FROM cell AS C WHERE '
                     +addQs(['C.%s'%x for x in cell_cols],' AND '))
             ,args = cell_args)

       ,Block(name = 'insert_struct'
             ,text = """INSERT INTO struct (raw,rawhash,cell_id)
                           VALUES (%s,SHA2(%s,512),%s)
                        ON DUPLICATE KEY UPDATE cell_id=cell_id"""
             ,args = [Arg('structs_few')
                      ,Arg('structs_few')
                     ,Arg('query_cell',0)])

       ,Block(name = 'query_struct_few'
             ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
             ,args = [Arg('structs_few')])

       ,Block(name = 'query_struct_many'
             ,text = 'SELECT S.struct_id FROM struct AS S WHERE S.rawhash = SHA2(%s,512)'
             ,args = [Arg('structs_many')])

       ,Block(name = 'insert_atom'
             ,text = mkInsCmd('atom',['struct_id','atom_id','element_id'
                                  ,'x','y','z','constrained','magmom','tag']
                            ,sqlite=False)
             ,args = [Arg('query_struct_many')]
                    +[Arg('atoms',i) for i in range(1,9)])

     ,Block(name = 'insert_traj'
           ,text = mkInsCmd('traj',['job_id','traj_id','struct_id','energy','final'],sqlite=False)
           ,args = [Arg('job_id')
                   ,Arg('get_traj',1)
                   ,Arg('query_struct_few')
                   ,Arg('get_traj',2)
                   ,Arg('get_traj',4)])

    ,Block(name = 'insert_trajatom'
          ,text = mkInsCmd('trajatom',['job_id','traj_id','struct_id','atom_id','fx','fy','fz'],sqlite=False)
          ,args = [Arg('job_id')
                  ,Arg('trajatoms',0)
                  ,Arg('query_struct_many')]
                 +[Arg('trajatoms',i) for i in range(1,5)])]))

allcols = ['element_id', 'symbol', 'name', 'atomic_weight','atomic_radius'
          , 'phase','evaporation_heat', 'pointgroup','spacegroup'
          , 'melting_point', 'metallic_radius', 'vdw_radius'
          , 'density', 'en_allen', 'is_radioactive'
          , 'lattice_struct' , 'fusion_heat'
          , 'econf', 'period', 'covalent_radius_bragg'
          , 'geochemical_class', 'abundance_crust', 'heat_of_formation'
          , 'electron_affinity', 'atomic_volume',  'boiling_point'
          , 'proton_affinity', 'covalent_radius_slater'
          , 'lattice_constant', 'dipole_polarizability'
          , 'en_ghosh', 'thermal_conductivity', 'group_id', 'en_pauling'
          , 'gas_basicity','abundance_sea']

populate_element = Rule(name='element'
    ,template = SimpleTemp(parse_mendeleev,[],allcols)
    ,metatemp = SimpleInsert('element',allcols))

sys_type = Rule(name='systype'
    ,query      = """SELECT S.struct_id,S.raw
                     FROM struct AS S
                     WHERE S.system_type IS NULL"""
    ,template   = Temp(tempfuncs=[TempFunc(get_system_type
                                          ,name = 'system_type'
                                          ,args = [Arg('json_to_traj')])
                                 ,TempFunc(json_to_traj
                                          ,args = [Arg('raw')])])
    ,metatemp   = SimpleUpdate(tab  = struct
                              ,cols = ['system_type']))


cell_info = Rule(name='cell'
    ,metatemp=PureSQL(["""UPDATE cell AS C SET surface_area = SQRT((C.a0*C.b2-C.a2*C.b1) * (C.a1*C.b2 - C.a2*C.b1)
                                                                 +(C.a0*C.b2-C.a2*C.b0) * (C.a0*C.b2 - C.a2*C.b0)
                                                                 +(C.a0*C.b1-C.a1*C.b0) * (C.a0*C.b1 - C.a1*C.b0))
                                              ,volume = C.a0*(C.b1*C.c2-C.b2*C.c1)
                                                       -C.a1*(C.b0*C.c2-C.b2*C.c0)
                                                       +C.a2*(C.b0*C.c1-C.b1*C.c0)
                                              ,a = SQRT(C.a0*C.a0+C.a1*C.a1+C.a2*C.a2)
                                              ,b = SQRT(C.b0*C.b0+C.b1*C.b1+C.b2*C.b2)
                                              ,c = SQRT(C.c0*C.c0+C.c1*C.c1+C.c2*C.c2)
                        WHERE C.cell_id > 0"""]))
n_atoms = Rule(name='n_atoms'
    ,metatemp=PureSQL(["""UPDATE struct SET n_atoms = (SELECT count(1)
                                FROM atom WHERE atom.struct_id=struct.struct_id
                                GROUP BY atom.struct_id)
                          WHERE struct_id > 0"""

                    ,"""UPDATE struct SET n_elems = (SELECT count(1)
                                    FROM struct_composition AS C WHERE C.struct_id=struct.struct_id
                                    GROUP BY C.struct_id)
                                    WHERE struct_id > 0"""]))

with open(environ['DFT_LOGFILES'],'r') as f:
    rootdata  = [l.split() for l in f.readlines() if '#' not in l and l.strip()!='']

pop_storage= Rule('root'
    ,metatemp=PureSQL(["""INSERT INTO roots (root,code,label,active) VALUES ('%s','%s','%s',1)
                        ON DUPLICATE KEY UPDATE storage_id=storage_id
                        """%(pth,code,label) for pth,code,label in rootdata]))

rules = [pop_storage
        ,populate_gpaw_job
        ,populate_qe_job
        ,populate_catalog_job
        ,populate_chgmol_job
        ,getJobMetadata
        ,populate_relax_calc
        ,populate_relax_job
        ,populate_traj
        ,populate_element
        ,sys_type
        ,cell_info
        ,n_atoms
        ,anyTraj]



from typing import Any, List, Callable, TypeVar, Dict,Iterable
A = TypeVar('A')
B = TypeVar('B')

################################################################################


def gcd(args: List[Any])->Any:
    """
    Greatest common denominator of a list
    """
    if len(args) == 1:
        return args[0]
    L = list(args)
    while len(L) > 1:
        a, b = L[len(L) - 2], L[len(L) - 1]
        L = L[:len(L) - 2]
        while a:
            a, b = b % a, a
        L.append(b)
    return abs(b)


def flatten(lol: List[List[A]])->List[A]:
    """Convert list of lists to a single list via concatenation"""
    return [item for sublist in lol for item in sublist]


def merge_dicts(dicts: List[Dict[A, B]])->Dict[A, B]:
    return {k: v for d in dicts for k, v in d.items()}


def concat_map(f: Callable[[A], List[B]], args: List[A]
               ) -> List[B]:
    """
    Maps a function over an input.
    We apply the function to every element in the list and concatenate result.

    """
    return flatten([f(arg) for arg in args])


def normalize_list(l: list) -> list:
    """
    [a,a,a,a,b,b,c,c] => [a,a,b,c]
    """
    if len(l) == 0:
        return l
    d = {x: l.count(x) for x in l}
    div = gcd(list(d.values()))
    norm = [[k] * (v // div) for k, v in d.items()]
    return [item for sublist in norm for item in sublist]



from typing import Any, List

from fractions import Fraction, gcd
from functools import reduce

# Numeric functions
# ------------------


def lcm(a: int, b: int)->int:
    return a * b // gcd(a, b)


def common_integer(numbers: List[float])->List[int]:
    fractions = [Fraction(n).limit_denominator() for n in numbers]
    multiple = reduce(lcm, [f.denominator for f in fractions])
    ints = [f * multiple for f in fractions]
    divisor = reduce(gcd, ints)
    return [int(n / divisor) for n in ints]


def roundfloat(x: Any) -> float:
    output = round(float(x), 3)
    return abs(output) if output == 0 else output


def safe_div(x: float, y: float)->float:
    try:
        return x / y
    except ZeroDivisionError:
        return 0.0



# External Modules
from typing     import List
from sqlite3    import connect as connect_lite
from pprint     import pformat,pprint
from time       import sleep
from sql        import Table,Join                   # type: ignore
from MySQLdb    import Warning,OperationalError     # type: ignore
from networkx   import Graph,NetworkXUnfeasible     # type: ignore
from networkx.algorithms import lexicographical_topological_sort,simple_cycles # type: ignore

from warnings import filterwarnings
filterwarnings("ignore", category = Warning)

# Internal Modules
from dbgen.support.datatypes.misc import ConnectInfo

"""
Tools for creating a database containing instructions for DBGen
"""

##############################################################################
# Interface with DB
#------------------
def sqlite_execute(db_path    : str
                  ,sqlcommand : str
                  ,binds      : list = []
                  ) -> None:
    """
    Execute a SQL statement that modifies the database
    """
    assert sqlcommand.lower()[:6] != 'select'
    #print(sqlcommand)
    with connect_lite(db_path) as conn:
        cursor = conn.cursor()
        cursor.execute('PRAGMA foreign_keys = ON')
        cursor.execute(sqlcommand, binds)
        output_array = cursor.fetchall()
        conn.commit()

def sqlite_select(db_path    : str
                 ,sqlcommand : str
                 ,binds      : list = []
                 ,mk_dict    : bool = False
                 ) -> list :
    """
    Select and return a list of tuples
    """
    def dict_factory(cursor, row):  # type: ignore
        return {col[0] : row[idx] for idx, col in enumerate(cursor.description)}

    assert sqlcommand.lower()[:6] == 'select'

    with connect_lite(db_path) as conn:
        cursor = conn.cursor()

        if mk_dict:
            cursor.row_factory = dict_factory

        cursor.execute(sqlcommand, binds)
        output_array = cursor.fetchall()
    return output_array

# Shortcuts
###########

def mkInsCmd(tabName : str
            ,names   : List[str]
            ,sqlite  : bool = False
            ) -> str:
    delim = ['?'] if sqlite else ['%s']
    dup    = ' ' if sqlite else ' ON DUPLICATE KEY UPDATE {0}={0}'.format(names[0])
    ins_names = names if sqlite else ['%s.%s'%(tabName,n) for n in names]
    return "INSERT INTO %s (%s) VALUES (%s) %s"%(tabName
                                             ,','.join(ins_names)
                                             ,','.join(delim*len(names))
                                             ,dup)
def mkUpdateCmd(tabName  : str
                ,names   : List[str]
                ,keys    : List[str]
                ) -> str:
    return "UPDATE %s SET %s WHERE %s"%(tabName
                                             ,addQs(names,',')
                                             ,addQs(keys,' AND '))

namespaceCmd = mkInsCmd('namespace',['name','type'],sqlite=True)

def handle_to_ind(pth:str,name:str)->int:
    cmd = 'SELECT rule_id FROM rule WHERE name = ?'
    try:
        return sqlite_select(pth,cmd,[name])[0][0]
    except IndexError:
        raise ValueError('Could not find rule '+name)

def func_to_ind(pth:str,name:str)->int:
    cmd = 'SELECT func_id FROM func WHERE name = ?'
    try:
        return sqlite_select(pth,cmd,[name])[0][0]
    except IndexError:
        raise ValueError('Could not find function '+name)

def get_tab_id(pth:str,name:str)->int:
    cmd = 'SELECT tab_id FROM tab WHERE name = ?'
    try:
        return sqlite_select(pth,cmd,[name])[0][0]
    except IndexError:
        raise ValueError('Could not find table '+name)
##############################################################################

def select_dict(conn  : ConnectInfo
               ,q     : str
               ,binds : list = []
               ) -> List[dict]:
    print('SELECTING with: ',q)
    with conn.mk_conn(mk_dict=True) as cxn: # type: ignore
        if 'group_concat' in q.lower():
            cxn.execute("SET SESSION group_concat_max_len = 1000000")
        cxn.execute(q,args=binds)
        return cxn.fetchall()

def sqlselect(conn  : ConnectInfo
             ,q     : str
             ,binds : list = []
             ) -> List[tuple]:
    #print('\n\nSQLSELECT ',q)#,binds)
    with conn.mk_conn() as cxn: # type: ignore
        if 'group_concat' in q.lower():
            cxn.execute("SET SESSION group_concat_max_len = 100000")
        cxn.execute(q,args=binds)
        return cxn.fetchall()

def sqlexecute(conn  : ConnectInfo
              ,q     : str
              ,binds : list = []
              ) -> None:
    #print('\n\nSQLEXECUTE ',q)#,binds)
    with conn.mk_conn() as cxn: # type: ignore
        cxn.execute("SET SESSION auto_increment_offset = 1")
        cxn.execute("SET SESSION auto_increment_increment = 1")
        while True:
            try:
                cxn.execute(q,args=binds)
                break
            except OperationalError as e:
                if  e.args[0] not in [1205,1213]: # deadlock error codes
                    sleep(10)
                    raise OperationalError(e)

def sqlexecutemany(conn  : ConnectInfo
                  ,q     : str
                  ,binds : List[list]
                  ) -> None:
    #print('executemany Q: ', q,binds)
    with conn.mk_conn() as cxn: # type: ignore
        cxn.execute("SET SESSION auto_increment_offset = 1")
        cxn.execute("SET SESSION auto_increment_increment = 1")

        while True:
            try:
                cxn.executemany(q,args=binds)
                break
            except OperationalError as e:
                if  e.args[0] not in [1205,1213]: # deadlock error codes
                    sleep(10)
                    raise OperationalError(e)
# ##############################################################################
# String shortcuts
#----------------
def addQs(xs:list,delim:str)->str:
    """ Ex: ['a','b','c'] + ',' ==> 'a = %s, b = %s, c = %s' """
    return delim.join(['{0} = %s'.format(x) for x in xs])

def topsort_with_dict(G:Graph,d:dict)->List:
    try:
        sortd = list(lexicographical_topological_sort(G))
        return [d[x] for x in sortd]
    except NetworkXUnfeasible:
        pprint(list(simple_cycles(G)))
        raise ValueError("(Cycle found)")
